import enum
import logging
import threading
import time
from ctypes import *
from ctypes.wintypes import *
from typing import Callable

from .vk import Vk
from .winevent import HWINEVENTHOOK, WINEVENTHOOKPROC, WinEvent

logger = logging.getLogger(__name__)

user32 = WinDLL("user32", use_last_error=True)

HC_ACTION = 0
# incoming message
WM_QUIT = 0x0012


# types for the hook including input parameter and return result

ULONG_PTR = WPARAM
LRESULT = LPARAM
# LPMSG = POINTER(MSG)

# Hook
HOOKPROC = WINFUNCTYPE(LRESULT, c_int, WPARAM, LPARAM)


# setup api


def _errcheck_bool(result, func, args):
    if not result:
        raise WinError(get_last_error())
    return args


user32.SetWindowsHookExW.errcheck = _errcheck_bool
user32.SetWindowsHookExW.restype = HHOOK
user32.SetWindowsHookExW.argtypes = (
    c_int,  # _In_ idHook
    HOOKPROC,  # _In_ lpfn
    HINSTANCE,  # _In_ hMod
    DWORD,
)  # _In_ dwThreadId

user32.CallNextHookEx.restype = LRESULT
user32.CallNextHookEx.argtypes = (
    HHOOK,  # _In_opt_ hhk
    c_int,  # _In_     nCode
    WPARAM,  # _In_     wParam
    LPARAM,
)  # _In_     lParam

user32.GetMessageW.argtypes = (
    LPMSG,  # _Out_    lpMsg
    HWND,  # _In_opt_ hWnd
    UINT,  # _In_     wMsgFilterMin
    UINT,
)  # _In_     wMsgFilterMax

user32.TranslateMessage.argtypes = (LPMSG,)
user32.DispatchMessageW.argtypes = (LPMSG,)

user32.SetWinEventHook.restype = HWINEVENTHOOK

# keyboard hook definition


class KBDLLHOOKMSGID(enum.IntEnum):
    """Keyboard event msgid"""

    WM_KEYDOWN = 0x0100
    WM_KEYUP = 0x0101
    WM_SYSKEYDOWN = 0x0104
    WM_SYSKEYUP = 0x0105


class KBDLLHOOKDATA(Structure):
    """Keyboard event data"""

    vkCode: int
    scanCode: int
    flags: int
    time: int
    dwExtraInfo: int

    _fields_ = (
        ("vkCode", DWORD),
        ("scanCode", DWORD),
        ("flags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", ULONG_PTR),
    )


# mouse hook definition


class MSLLHOOKMSGID(enum.IntEnum):
    """Mouse event msgid

    Ref: https://learn.microsoft.com/en-us/windows/win32/inputdev/wm-xbuttondown
    """

    WM_MOUSEMOVE = 0x0200
    WM_LBUTTONDOWN = 0x0201
    WM_LBUTTONUP = 0x0202
    WM_RBUTTONDOWN = 0x0204
    WM_RBUTTONUP = 0x0205
    WM_MBUTTONDOWN = 0x0207
    WM_MBUTTONUP = 0x0208
    WM_MOUSEWHEEL = 0x020A
    WM_MOUSEHWHEEL = 0x020E
    WM_XBUTTONDOWN = 0x020B
    WM_XBUTTONUP = 0x020C


class MSLLHOOKDATA(Structure):
    """Mouse event data"""

    pt: POINT
    mouseData: int
    flags: int
    time: int
    dwExtraInfo: int

    _fields_ = (
        ("pt", POINT),
        ("mouseData", DWORD),
        ("flags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", ULONG_PTR),
    )

    def get_wheel_delta(self) -> int:
        """Get mouse wheel delta

        Ref: https://docs.microsoft.com/en-us/windows/win32/inputdev/wm-mousewheel
        """
        return SHORT(self.mouseData >> 16).value

    def hiword(self) -> int:
        return WORD(self.mouseData >> 16).value

    def loword(self) -> int:
        return WORD(self.mouseData & 0xFFFF).value


_hooks = {}


def hook(hook_id, wparam_type, lparam_type, cb) -> HANDLE:
    # for the hooks to work, note that only low level keyboard/mouse work this way
    # while others require DLL injection
    @HOOKPROC
    def proc(nCode, wParam, lParam):
        if nCode == HC_ACTION:
            wparam = wparam_type(wParam)
            lparam = cast(lParam, POINTER(lparam_type))[0]
            if cb(wparam, lparam):
                return 1
        return user32.CallNextHookEx(None, nCode, wParam, lParam)

    hook = user32.SetWindowsHookExW(hook_id, proc, None, 0)
    # keep a reference to the callback to prevent it from being garbage collected
    global _hooks
    _hooks[hook] = proc
    return hook


def unhook(hook: HANDLE):
    """Unhook"""
    global _hooks
    if hook in _hooks:
        _hooks.pop(hook)
        user32.UnhookWindowsHookEx(hook)


_winevent_hooks = {}


def hook_winevent(
    event_min: WinEvent,
    event_max: WinEvent,
    cb: Callable[[WinEvent, HWND, LONG, LONG, DWORD, DWORD], None],
) -> HANDLE:
    """Hook window events

    Usage:

    .. code-block:: python
        def winevent_callback(
            event: WinEvent,
            hwnd: HWND,
            id_obj: LONG,
            id_chd: LONG,
            id_evt_thread: DWORD,
            time: DWORD,
        ):
            pass

        hook_winevent(winevent_callback)
    """

    @WINEVENTHOOKPROC
    def proc(hhook, event, hwnd, id_obj, id_chd, id_evt_thread, dwms_evt_time):
        cb(WinEvent(event), hwnd, id_obj, id_chd, id_evt_thread, dwms_evt_time)

    hook = user32.SetWinEventHook(
        DWORD(event_min.value),
        DWORD(event_max.value),
        HMODULE(None),
        proc,
        0,
        0,
        0,  # WINEVENT_OUTOFCONTEXT
    )
    # keep a reference to the callback to prevent it from being garbage collected
    global _hooks
    _winevent_hooks[hook] = proc
    return hook


def unhook_winevent(hook: HANDLE):
    """Unhook"""
    global _winevent_hooks
    if hook in _winevent_hooks:
        _winevent_hooks.pop(hook)
        user32.UnhookWinEvent(hook)


def hook_keyboard(cb: Callable[[KBDLLHOOKMSGID, KBDLLHOOKDATA], bool]) -> HANDLE:
    """Install keyboard hook

    Usage:

    .. code-block:: python

        def swallow_keyboard_a_key_event(msgid: KBDLLHOOKMSGID, data: KBDLLHOOKDATA) -> bool:
            return data.vkCode == VirtualKey.VK_A:

        hook_keyboard(swallow_keyboard_a_key_event)

    :param callback: function to be called when key press/release, return ``True`` to stop
                     propagation
    :return: hook handle (for unhook) and callback function(must be reference somewhere or it will be GCed),
    :rtype: Tuple[HANDLE, Callable]
    """
    return hook(13, KBDLLHOOKMSGID, KBDLLHOOKDATA, cb)


def hook_mouse(cb: Callable[[MSLLHOOKMSGID, MSLLHOOKDATA], bool]) -> HANDLE:
    """Install mouse hook

    Usage:

    .. code-block:: python

        def swallow_mouse_middle_btn_event(msgid: MSLLHOOKMSGID, data: MSLLHOOKDATA) -> bool:
            return msgid == MSLLHOOKMSGID.WM_MBUTTONUP or msgid == MSLLHOOKMSGID.WM_MBUTTONDOWN:

        hook_mouse(swallow_mouse_middle_btn_event)

    :param callback: function to be called when mouse moved, bth pressed/release and scroll,
                        return ``True`` to stop propagation
    :return: hook handle (for unhook) and callback function(must be reference somewhere or it will be GCed),
    :rtype: Tuple[HANDLE, Callable]
    """
    return hook(14, MSLLHOOKMSGID, MSLLHOOKDATA, cb)


def message_loop():
    """For debugging purpose"""
    import signal

    signal.signal(signal.SIGINT, signal.SIG_DFL)
    msg = byref(MSG())
    while True:
        bRet = user32.GetMessageW(msg, None, 0, 0)
        if not bRet:
            break
        if bRet == -1:
            raise WinError(get_last_error())
        user32.TranslateMessage(msg)
        user32.DispatchMessageW(msg)


if __name__ == "__main__":
    import sys
    from datetime import datetime

    from . import window

    logging.basicConfig(level=logging.DEBUG)

    def keyboard(msgid: KBDLLHOOKMSGID, msg: KBDLLHOOKDATA) -> bool:
        print(
            "{:15s} {:15s}: vkCode {:3x} scanCode {:3x} flags: {:3d}, time: {} extra: {}".format(
                Vk(msg.vkCode).name,
                msgid.name,
                msg.vkCode,
                msg.scanCode,
                msg.flags,
                msg.time,
                msg.dwExtraInfo,
            )
        )

    def mouse(msgid: MSLLHOOKMSGID, msg: MSLLHOOKDATA) -> bool:
        print(
            "{:15s}  x: {:3d} y: {:3d} hi: {:5x} lo: {:5x} flags: {:3x} extra: {:6x} t: {:d}".format(
                msgid.name,
                msg.pt.x,
                msg.pt.y,
                int(msg.hiword()),
                int(msg.loword()),
                msg.flags,
                msg.dwExtraInfo,
                msg.time,
            )
        )
        if msgid == MSLLHOOKMSGID.WM_MOUSEWHEEL:
            print("delta: {}".format(msg.get_wheel_delta()))

    def winevent(
        event: WinEvent,
        hwnd: HWND,
        id_obj: LONG,
        id_chd: LONG,
        id_evt_thread: DWORD,
        time: DWORD,
    ):
        print("==================================")
        print(
            "[{now}] {event:30s} {hwnd:8d} ido: {id_obj:6d} idc: {id_chd:6d} {title}".format(
                now=datetime.now().strftime("%M:%S.%f"),
                event=event.name,
                hwnd=hwnd or 0,
                id_obj=id_obj,
                id_chd=id_chd,
                title=window.Window(hwnd).title,
            )
        )
        window.inspect_window(hwnd)
        print("==================================")

    kb_hook = hook_keyboard(keyboard)
    ms_hook = hook_mouse(mouse)
    we_hook = hook_winevent(
        WinEvent.EVENT_SYSTEM_MINIMIZESTART, WinEvent.EVENT_SYSTEM_MINIMIZEEND, winevent
    )

    def unhook_mouse_after_10s():
        time.sleep(10)
        unhook(ms_hook)
        print("mouse unhooked")

    threading.Thread(target=unhook_mouse_after_10s).start()

    message_loop()



import os
from configparser import ConfigParser
from datetime import date
from typing import Optional

DEFAULT_STATE_PATH = os.path.join(os.getenv("LOCALAPPDATA"), "jigsawwm", "state.json")


class StateManager:
    config: ConfigParser
    state_path: str

    def __init__(self, state_path: str = DEFAULT_STATE_PATH):
        self.config = ConfigParser()
        self.state_path = state_path
        self.load()

    def load(self):
        if os.path.exists(self.state_path):
            self.config.read(self.state_path, encoding="utf-8")

    def save(self):
        if not os.path.exists(self.state_path):
            os.makedirs(os.path.dirname(self.state_path), exist_ok=True)
        with open(self.state_path, "w+", encoding="utf8") as f:
            self.config.write(f)

    def get(self, section: str, option: str, default: str = None) -> str:
        if not self.config.has_section(section):
            self.config.add_section(section)
        return self.config.get(section, option, fallback=default)

    def set(self, section: str, option: str, value: str = None):
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, option, value)

    def getbool(self, section: str, option: str, default: bool = False) -> bool:
        return bool(self.get(section, option, default))

    def setbool(self, section: str, option: str, value: bool):
        self.set(section, option, str(value).lower())

    def getdate(
        self, section: str, option: str, default: date = None
    ) -> Optional[date]:
        d = self.get(section, option, default)
        if d:
            return date.fromisoformat(d)
        return None

    def setdate(self, section: str, option: str, vaule: date = None):
        self.set(section, option, vaule.isoformat() if date else None)


_state_manager = None


def set_state_path(state_path: str):
    global _state_manager
    _state_manager = StateManager(state_path)


def get_state_manager():
    global _state_manager
    if not _state_manager:
        _state_manager = StateManager()
    return _state_manager



from ctypes import *
from ctypes.wintypes import *

from comtypes import *

# Ref: https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nn-shobjidl_core-ivirtualdesktopmanager


def _check(hresult: HRESULT):
    if hresult:
        raise Exception(f"HRESULT: {hresult}")


class IVirtualDesktopManager(IUnknown):
    _iid_ = GUID("{A5CD92FF-29BE-454C-8D04-D82879FB3F1B}")

    _methods_ = [
        COMMETHOD(
            [],
            HRESULT,
            "IsWindowOnCurrentVirtualDesktop",
            (["in"], HWND, "topLevelWindow"),
            (["out"], LPBOOL, "onCurrentDesktop"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "GetWindowDesktopId",
            (["in"], HWND, "topLevelWindow"),
            (["out"], POINTER(GUID), "desktopId"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "MoveWindowToDesktop",
            (["in"], HWND, "topLevelWindow"),
            (["in"], POINTER(GUID), "desktopId"),
        ),
    ]

    def GetWindowDesktopId(self, hwnd: HWND) -> GUID:
        """Retrieves virtual desktop id of the specified window handle,
        it works only when ``hwnd`` is NOT elevated (Administrator)

        :param HWND hwnd: window handle
        :returns: desktop id
        :rtype: comtypes.GUID
        """
        desktop_id = GUID()
        _check(self.__com_GetWindowDesktopId(hwnd, pointer(desktop_id)))
        return desktop_id

    def IsWindowOnCurrentVirtualDesktop(self, hwnd: HWND) -> bool:
        """Checks if ``hwnd`` belongs to the current desktop
        it works only when ``hwnd`` is NOT elevated (Administrator)

        :param HWND hwnd: window handle
        :returns: True if window belongs to current virtual desktop
        :rtype: bool
        """
        value = BOOL()
        _check(self.__com_IsWindowOnCurrentVirtualDesktop(hwnd, pointer(value)))
        return value.value

    def MoveWindowToDesktop(self, hwnd: HWND, desktop_id: GUID):
        """Moves window to desktop
        it works only when ``hwnd`` is NOT elevated (Administrator)

        :param HWND hwnd: window handle
        :param GUID desktop_id: target virtual desktop id
        """
        _check(self.__com_MoveWindowToDesktop(hwnd, pointer(desktop_id)))


virtual_desktop_manager: IVirtualDesktopManager = CoCreateInstance(
    GUID("{AA509086-5CA9-4C25-8F95-589D3C07B48A}"), interface=IVirtualDesktopManager
)


if __name__ == "__main__":
    import time

    time.sleep(2)
    virtual_desktop_manager.GetWindowDesktopId(windll.user32.GetForegroundWindow())
    # print(virtual_desktop_manager.IsWindowOnCurrentVirtualDesktop(HWND(131352)))



from dataclasses import dataclass
from datetime import datetime
from os import path
from typing import Dict, List, Optional, Set

from jigsawwm.tiler.tilers import *
from jigsawwm.virtdeskstub import find_or_create_virtdeskstub
from jigsawwm.w32 import hook
from jigsawwm.w32.idesktopwallpaper import desktop_wallpaper
from jigsawwm.w32.ivirtualdesktopmanager import GUID, virtual_desktop_manager
from jigsawwm.w32.monitor import (
    Monitor,
    get_monitor_from_cursor,
    get_monitor_from_window,
    get_monitors,
    get_topo_sorted_monitors,
    set_cursor_pos,
)
from jigsawwm.w32.window import (
    DWORD,
    HWND,
    LONG,
    RECT,
    Window,
    get_active_window,
    get_first_app_window,
    get_foreground_window,
    get_manageable_windows,
    is_app_window,
    is_window,
    sprint_window,
)
from jigsawwm.w32.winevent import WinEvent


@dataclass
class Theme:
    """Theme is a set of preference packed together for users to switch easily,
    typically, it consists of a LayoutTiler, Background, Gap between windows and
    other options.
    """

    # name of the theme
    name: str
    # layout tiler
    layout_tiler: LayoutTiler
    # unused
    icon_name: Optional[str] = None
    # unused
    icon_path: Optional[str] = None
    # background, color if the string starts with `#`, otherwise treated as image path
    background: Optional[str] = None
    # new appeared window would be prepended to the list if the option was set to True
    new_window_as_master: Optional[bool] = None
    # gap between windows / monitor edges
    gap: Optional[int] = 0
    # forbid
    strict: Optional[bool] = None
    hook_ids: List[int] = None


class MonitorState:
    """MonitorState holds variables needed by a Monitor


    :param VirtDeskState virtdesk_state: associated virtual desktop
    :param Monitor monitor: associated system monitor
    :param str theme: the active theme for the monitor in the virtual desktop
    """

    virtdesk_state: "VirtDeskState"
    monitor: Monitor
    theme: Optional[str]
    windows: List[Window]
    last_active_window: Optional[Window] = None

    def __init__(
        self,
        virtdesk_state: "VirtDeskState",
        monitor: Monitor,
        theme: Optional[str] = None,
    ):
        self.virtdesk_state = virtdesk_state
        self.monitor = monitor
        self.theme = theme
        self.windows = []
        self.last_active_window = None
        self.hook_ids = []

    def get_theme(self) -> Theme:
        """Retrieves theme for monitor in current virtual desktop"""
        mgr = self.virtdesk_state.manager
        return mgr.themes[mgr.get_theme_index(self.theme)]

    def get_existing_windows(self) -> List[Window]:
        """Retrieves current managed windows"""
        return self.windows

    def sync(self, windows: Set[Window], restrict=False):
        """Synchronize managed windows with given actual windows currently visible and arrange them accordingly

        :param Set[Window] windows: latest visible windows
        :param bool restrict: optional, restrict windows to their specified rect no matter what
        """
        theme = self.get_theme()
        old_list = self.windows
        new_list = []
        for w in old_list:
            if w not in windows:
                continue
            windows.remove(w)
            if not w.exists():
                continue
            new_list.append(w)
        # and then, prepend or append the new windows
        if theme.new_window_as_master:
            new_list = list(windows) + new_list
        else:
            new_list = new_list + list(windows)
        # skip if there is nothing changed, unless Strict mode is enable
        if new_list == old_list:
            if theme.background is not None:
                self.set_background(theme)
            if restrict:
                self.restrict(theme)
            return
        self.windows = new_list
        self.arrange(theme)

    def arrange(self, theme: Optional[Theme] = None):
        """Arrange windows based on the theme

        :param str theme: optional, fallback to theme of the instance
        """
        theme = theme or self.get_theme()
        if theme.background is not None:
            self.set_background(theme)
        scale_factor = self.monitor.get_scale_factor().value / 100
        wr = self.monitor.get_info().rcWork
        work_area = (wr.left, wr.top, wr.right, wr.bottom)
        windows = self.get_existing_windows()
        i = 0
        gap = theme.gap
        # print(work_area, theme.name, len(windows))
        for left, top, right, bottom in theme.layout_tiler(work_area, len(windows)):
            window = windows[i]
            # add gap
            if gap:
                if left == wr.left:
                    left += gap
                if top == wr.top:
                    top += gap
                if right == wr.right:
                    right -= gap
                if bottom == wr.bottom:
                    bottom -= gap
            left += gap
            top += gap
            right -= gap
            bottom -= gap
            # compensation
            r = window.get_rect()
            b = window.get_extended_frame_bounds()
            # now, here is the tricky part, thanks to the chaotic Win32 API
            #
            #   1. the `SetWindowPos` api accept Scaled-Pixel (x,y,w,h), however,
            #      most of windows would be rendered smaller than the given value.
            #   2. ideally, we want windows to be rendered in exact size we specified.
            #      the only way I found is to get the actual size by `dwmapi` and
            #      compensate the difference. however `DwmGetWindowAttribute` works in
            #      Physical-Pixel, it must be scaled
            compensated_rect = (
                round(left + r.left - (b.left / scale_factor)),
                round(top + r.top - (b.top / scale_factor)),
                round(right + r.right - (b.right / scale_factor)),
                round(bottom + r.bottom - (b.bottom / scale_factor)),
            )
            window.set_rect(RECT(*compensated_rect))
            i += 1

    def set_background(self, theme: Theme):
        """Set background for the monitor based on given theme"""
        monitors = list(get_monitors())
        idx = monitors.index(self.monitor)
        monitor_id = desktop_wallpaper.GetMonitorDevicePathAt(idx)
        desktop_wallpaper.SetWallpaper(monitor_id, theme.background)

    def restrict(self, theme: Optional[Theme] = None):
        """Restrict all managed windows to their specified rect"""
        theme = theme or self.get_theme()
        if not theme.strict:
            return
        for window in self.windows:
            window.set_rect(window.last_rect)


class VirtDeskState:
    """VirtDeskState holds variables needed by a Virtual Desktop

    :param WindowManager manager: associated WindowManager
    :param GUID desktop_id: virtual desktop id
    """

    desktop_id: GUID
    manager: "WindowManager"
    managed_windows: Set[Window]
    monitors: Dict[Monitor, MonitorState]
    last_active_window: Optional[Window] = None

    def __init__(self, manager: "WindowManager", desktop_id: GUID):
        self.desktop_id = desktop_id
        self.manager = manager
        self.managed_windows = set()
        self.monitors = {}
        self.last_active_window = None

    def get_monitor(self, monitor: Monitor) -> MonitorState:
        """Retrieves the monitor state for the specified monitor in the virtual desktop

        :param Monitor monitor: monitor
        :returns: monitor state
        :rtype: MonitorState
        """
        monitor_state = self.monitors.get(monitor)
        if monitor_state is None:
            monitor_state = MonitorState(self, monitor)
            self.monitors[monitor] = monitor_state
        return monitor_state

    def get_managed_active_window(self) -> Optional[Window]:
        """Retrieves the managed forground window if any"""
        window = get_active_window()
        if window is None:
            return None
        if window not in self.managed_windows:
            return None
        return window

    def get_last_managed_active_window(self) -> Optional[Window]:
        """Retrieves the latest managed forground window if any"""
        if self.last_active_window and (
            self.last_active_window not in self.managed_windows
            or not self.last_active_window.exists()
        ):
            return None
        return self.last_active_window

    def find_owner(self, window: Window) -> Optional[MonitorState]:
        """Retrieves the windows list containing specified window and its index in the list"""
        monitor = get_monitor_from_window(window.handle)
        monitor_state = self.get_monitor(monitor)
        if window in monitor_state.windows:
            return monitor_state
        return None


class WindowManager:
    """WindowManager detect the monitors/windows state and arrange them dynamically
    keep in mind that not all windows are managed by the WindowManager, only those
    returned by `jigsawwm.w32.get_normal_windows()` and not ignored would be managed.

    The WindowManager works just like other Dynamic Window Managers, it store all
    managed windows in a list, the first one in the list is called Master, it would
    normally take up most area of the screen while others occuppy the rest.

    :param List[Theme] themes: all avaiable themes for user to switch
    :param ignore_exe_names: list of executable filenames that you don't want them
                             to be managed/arranged
    """

    _state: Dict[GUID, VirtDeskState]
    themes: List[Theme]
    ignore_exe_names: Set[str]
    force_managed_exe_names: Set[str]

    def __init__(
        self,
        themes: List[Theme] = None,
        ignore_exe_names: Set[str] = None,
        force_managed_exe_names: Set[str] = None,
    ):
        self._state = {}
        self.themes = themes or [
            Theme(
                name="WideScreen Dwindle",
                layout_tiler=widescreen_dwindle_layout_tiler,
                icon_name="wide-dwindle.png",
            ),
            Theme(
                name="OBS Dwindle",
                layout_tiler=obs_dwindle_layout_tiler,
                icon_name="obs.png",
            ),
            Theme(
                name="Dwindle",
                layout_tiler=dwindle_layout_tiler,
                icon_name="dwindle.png",
            ),
        ]
        self.ignore_exe_names = set(ignore_exe_names or [])
        self.force_managed_exe_names = set(force_managed_exe_names or [])
        self.theme = self.themes[0].name
        self.sync(init=True)

    def try_get_virtdesk_state(
        self, hwnd: Optional[HWND] = None
    ) -> Optional[VirtDeskState]:
        """Retrieve virtual desktop state without exception.
        It is likely to fail due to API limitation, i.e. pressing hotkey while
        Start Menu is activating, better to do nothing than raising exception"""
        try:
            return self.get_virtdesk_state(hwnd)
        except:
            pass

    def get_virtdesk_state(self, hwnd: Optional[HWND] = None) -> VirtDeskState:
        """Retrieve virtual desktop state"""
        # Hey, M$, why not just offer an API so we can know which virtual desktop is current active? WHY?
        proc = None
        # try to use the first app window
        if hwnd is None:
            hwnd = get_first_app_window()
        # last resort: create a temporary window ... :tears:
        if hwnd is None:
            hwnd, proc = find_or_create_virtdeskstub()

        desktop_id = None
        try:
            desktop_id = virtual_desktop_manager.GetWindowDesktopId(hwnd)
        except Exception as e:
            pass
        if desktop_id is None or desktop_id == GUID():
            wininfo = sprint_window(hwnd)
            raise Exception("invalid desktop_id\n" + wininfo)
        # print("desktop id", desktop_id)
        virtdesk_state = self._state.get(desktop_id)
        if virtdesk_state is None:
            # make sure monitor_state for current virtual desktop exists
            virtdesk_state = VirtDeskState(self, desktop_id)
            self._state[desktop_id] = virtdesk_state
        if proc:
            proc.kill()
        return virtdesk_state

    def is_ignored(self, window: Window) -> bool:
        exepath = window.exe
        return not exepath or path.basename(exepath) in self.ignore_exe_names

    def is_force_managed(self, hwnd: HWND) -> bool:
        try:
            exepath = Window(hwnd).exe
            return exepath and path.basename(exepath) in self.force_managed_exe_names
        except:
            return False

    def sync(self, init=False, restrict=False) -> bool:
        """Update manager state(monitors, windows) to match OS's and arrange windows if it is changed"""
        manageable_windows = list(get_manageable_windows(self.is_force_managed))
        if not manageable_windows:
            return
        virtdesk_state = self.try_get_virtdesk_state(manageable_windows[0].handle)
        if not virtdesk_state:
            return
        # gather all manageable windows and group them by monitor
        group_wins_by_mons: Dict[Monitor, Set[Window]] = {}
        managed_windows = set()
        for window in manageable_windows:
            # skip certain exe file name
            if self.is_ignored(window):
                continue
            # if window was already managed, use previous monitor, or use the one under the cursor
            if init or window in virtdesk_state.managed_windows:
                monitor = get_monitor_from_window(window.handle)
            else:
                monitor = get_monitor_from_cursor()
            windows = group_wins_by_mons.get(monitor)
            if windows is None:
                windows = set()
                group_wins_by_mons[monitor] = windows
            windows.add(window)
            managed_windows.add(window)
        virtdesk_state.managed_windows = managed_windows

        # pass down to monitor_state for further synchronization
        for monitor, windows in group_wins_by_mons.items():
            monitor_state = virtdesk_state.get_monitor(monitor)
            monitor_state.sync(windows, restrict=restrict)

    def arrange_all_monitors(self):
        """Arrange all windows in all monitors to where their suppose to be"""
        virtdesk_state = self.try_get_virtdesk_state()
        if not virtdesk_state:
            return
        for monitor in virtdesk_state.monitors.values():
            monitor.arrange()

    def activate(self, window: Window):
        """Activate specified window"""
        virtdesk_state = self.try_get_virtdesk_state(window.handle)
        if not virtdesk_state:
            return
        window.activate()
        # move cursor to the center of the window
        rect = window.get_rect()
        set_cursor_pos(
            rect.left + (rect.right - rect.left) / 2,
            rect.top + (rect.bottom - rect.top) / 2,
        )
        virtdesk_state.last_active_window = window
        monitor = get_monitor_from_window(window.handle)
        virtdesk_state.get_monitor(monitor).last_active_window = window

    def activate_by_offset(self, offset: int):
        """Activate managed window by offset

        When the active window is managed, activate window in the same monitor by offset
        When the active window is unmanaged, activate the last active window
        When none of above viable, activate the Master window in the montior under cursor
        Or, do nothing!
        """
        virtdesk_state = self.try_get_virtdesk_state()
        if not virtdesk_state:
            return
        src_window = virtdesk_state.get_managed_active_window()
        dst_window = None
        if src_window:
            monitor_state = virtdesk_state.find_owner(src_window)
            if len(monitor_state.windows) > 0:
                src_index = monitor_state.windows.index(src_window)
                dst_index = (src_index + offset) % len(monitor_state.windows)
                dst_window = monitor_state.windows[dst_index]
        else:
            dst_window = virtdesk_state.last_active_window
        if not dst_window:
            monitor_state = virtdesk_state.get_monitor(get_monitor_from_cursor())
            if len(monitor_state.windows) > 0:
                dst_window = monitor_state.windows[0]
        if not dst_window:
            return
        self.activate(dst_window)

    def activate_next(self):
        """Activate the managed window next to the last activated managed window"""
        self.activate_by_offset(+1)

    def activate_prev(self):
        """Activate the managed window prior to the last activated managed window"""
        self.activate_by_offset(-1)

    def _reorder(self, reorderer: Callable[[List[Window], int], None]):
        virtdesk_state = self.try_get_virtdesk_state()
        if not virtdesk_state:
            return
        src_window = virtdesk_state.get_managed_active_window()
        if not src_window:
            return
        monitor_state = virtdesk_state.find_owner(src_window)
        if monitor_state is None:
            return
        if len(monitor_state.windows) < 2:
            return
        reorderer(monitor_state.windows, monitor_state.windows.index(src_window))
        monitor_state.arrange()
        self.activate(src_window)

    def swap_by_offset(self, offset: int):
        """Swap current active managed window with its sibling by offset"""

        def reorderer(windows: List[Window], src_idx: int):
            dst_idx = (src_idx + offset) % len(windows)
            windows[src_idx], windows[dst_idx] = windows[dst_idx], windows[src_idx]

        self._reorder(reorderer)

    def swap_next(self):
        """Swap the current active managed window with its next in list"""
        self.swap_by_offset(+1)

    def swap_prev(self):
        """Swap the current active managed window with its previous in list"""
        self.swap_by_offset(-1)

    def swap_master(self):
        """Swap the current active managed window with the Master or the second window
        in the list if it is Master already
        """

        def reorderer(windows: List[Window], src_idx: int):
            if src_idx == 0:
                dst_idx = 1
            else:
                dst_idx = 0
            windows[src_idx], windows[dst_idx] = windows[dst_idx], windows[src_idx]

        self._reorder(reorderer)

    def set_master(self):
        """Set the active active managed window as the Master or the second window
        in the list if it is Master already
        """

        def reorderer(windows: List[Window], src_idx: int):
            src_window = windows[src_idx]
            if src_idx == 0:
                src_idx = 1
                src_window = windows[1]
            # shift elements from the beginning to the src_window
            for i in reversed(range(1, src_idx + 1)):
                windows[i] = windows[i - 1]
            # assign new master
            windows[0] = src_window

        self._reorder(reorderer)

    def get_theme_index(self, theme_name: str) -> int:
        """Retrieves the index of given theme name, useful to switching theme"""
        i = len(self.themes) - 1
        while i > 0:
            if self.themes[i].name == theme_name:
                return i
            i -= 1
        return i

    def switch_theme_by_offset(self, delta: int):
        """Switch theme by offset"""
        virtdesk_state = self.try_get_virtdesk_state()
        if not virtdesk_state:
            return
        monitor = (
            get_monitor_from_window(get_foreground_window())
            or get_monitor_from_cursor()
        )
        monitor_state = virtdesk_state.get_monitor(monitor)
        theme_index = self.get_theme_index(monitor_state.theme)
        new_theme_name = self.themes[(theme_index + delta) % len(self.themes)].name
        monitor_state.theme = new_theme_name
        monitor_state.arrange()

    def prev_theme(self):
        """Switch to previous theme in the themes list"""
        self.switch_theme_by_offset(-1)

    def next_theme(self):
        """Switch to next theme in the themes list"""
        self.switch_theme_by_offset(+1)

    def get_monitor_state_pair(
        self, delta: int, virtdesk_state: Optional[VirtDeskState] = None
    ) -> Tuple[MonitorState, MonitorState]:
        """Retrieves a pair of monitor_states, from cursor and its offset in the list"""
        virtdesk_state = virtdesk_state or self.try_get_virtdesk_state()
        if not virtdesk_state:
            return
        monitors = get_topo_sorted_monitors()
        src_monitor = get_monitor_from_cursor()
        src_idx = monitors.index(src_monitor)
        dst_idx = (src_idx + delta) % len(monitors)
        dst_monitor = monitors[dst_idx]
        dst_monitor_state = virtdesk_state.get_monitor(dst_monitor)
        src_monitor_state = virtdesk_state.get_monitor(src_monitor)
        return src_monitor_state, dst_monitor_state

    def switch_monitor_by_offset(self, delta: int):
        """Switch to another monitor by given offset"""
        _, dst_monitor_state = self.get_monitor_state_pair(delta)
        window = dst_monitor_state.last_active_window
        if window is None or not window.exists():
            windows = dst_monitor_state.get_existing_windows()
            if windows:
                window = windows[0]
        if window:
            self.activate(window)
        else:
            rect = dst_monitor_state.monitor.get_info().rcWork
            set_cursor_pos(
                rect.left + (rect.right - rect.left) / 2,
                rect.top + (rect.bottom - rect.top) / 2,
            )

    def prev_monitor(self):
        """Switch to previous monitor"""
        self.switch_monitor_by_offset(-1)

    def next_monitor(self):
        """Switch to next monitor"""
        self.switch_monitor_by_offset(+1)

    def move_to_monitor_by_offset(self, delta: int):
        """Move active window to another monitor by offset"""
        virtdesk_state = self.try_get_virtdesk_state()
        if not virtdesk_state:
            return
        window = virtdesk_state.get_managed_active_window()
        if not window:
            return
        src_monitor_state, dst_monitor_state = self.get_monitor_state_pair(
            delta, virtdesk_state
        )
        idx = src_monitor_state.windows.index(window)
        src_monitor_state.windows.remove(window)
        src_monitor_state.arrange()
        dst_monitor_state.windows.append(window)
        dst_monitor_state.arrange()
        src_win_len = len(src_monitor_state.windows)
        if src_win_len:
            next_window = src_monitor_state.windows[idx % src_win_len]
            self.activate(next_window)

    def move_to_prev_monitor(self):
        """Move active window to previous monitor"""
        self.move_to_monitor_by_offset(-1)

    def move_to_next_monitor(self):
        """Move active window to next monitor"""
        self.move_to_monitor_by_offset(+1)

    def _winevent_callback(
        self,
        event: WinEvent,
        hwnd: HWND,
        id_obj: LONG,
        id_chd: LONG,
        id_evt_thread: DWORD,
        evt_time: DWORD,
    ):
        if (
            id_obj
            or id_chd
            or not is_window(hwnd)
            or not is_app_window(hwnd)
            or self.is_ignored(Window(hwnd))
        ):
            return
        print(
            "[{now}] {event:30s} {hwnd:8d} {title}".format(
                now=datetime.now().strftime("%M:%S.%f"),
                event=event.name,
                hwnd=hwnd,
                #  ido: {id_obj:6d} idc: {id_chd:6d}
                # id_obj=id_obj,
                # id_chd=id_chd,
                title=Window(hwnd).title,
            )
        )
        self.sync(restrict=event == WinEvent.EVENT_SYSTEM_MOVESIZEEND)

    def install_hooks(self):
        """Install hooks for window events"""
        self.hook_ids = [
            hook.hook_winevent(
                WinEvent.EVENT_OBJECT_SHOW,
                WinEvent.EVENT_OBJECT_HIDE,
                self._winevent_callback,
            ),
            hook.hook_winevent(
                WinEvent.EVENT_OBJECT_CLOAKED,
                WinEvent.EVENT_OBJECT_UNCLOAKED,
                self._winevent_callback,
            ),
            hook.hook_winevent(
                WinEvent.EVENT_SYSTEM_MINIMIZESTART,
                WinEvent.EVENT_SYSTEM_MINIMIZEEND,
                self._winevent_callback,
            ),
            hook.hook_winevent(
                WinEvent.EVENT_SYSTEM_MOVESIZEEND,
                WinEvent.EVENT_SYSTEM_MOVESIZEEND,
                self._winevent_callback,
            ),
        ]

    def uninstall_hooks(self):
        for hook_id in self.hook_ids:
            hook.unhook_winevent(hook_id)
        self.hook_ids = []


if __name__ == "__main__":
    wm = WindowManager()
    wm.sync()
    print()
    wm.sync()



from jigsawwm.jmk.core import *
from jigsawwm.w32.vk import *


def test_jmk_core_map_key(mocker):
    # map key
    jmk_core = JmkCore(
        next_handler=mocker.Mock(),
        layers=[
            {
                Vk.A: JmkKey(Vk.B),
            }
        ],
    )
    jmk_core(JmkEvent(Vk.A, True))
    assert jmk_core.next_handler.call_count == 1
    assert jmk_core.next_handler.call_args[0][0] == JmkEvent(Vk.B, True)
    jmk_core(JmkEvent(Vk.A, False))
    assert jmk_core.next_handler.call_count == 2
    assert jmk_core.next_handler.call_args[0][0] == JmkEvent(Vk.B, False)
    # pass through
    jmk_core(JmkEvent(Vk.B, True))
    jmk_core(JmkEvent(Vk.B, False))
    assert jmk_core.next_handler.call_count == 4


def test_jmk_core_taphold(mocker):
    jmk_core = JmkCore(
        next_handler=mocker.Mock(),
        layers=[{Vk.A: JmkTapHold(hold=Vk.LMENU, tap=Vk.A, term=0.3)}],
    )
    # hold
    jmk_core(JmkEvent(Vk.A, True))
    time.sleep(0.4)
    assert jmk_core.next_handler.call_count == 1
    assert jmk_core.next_handler.call_args[0][0] == JmkEvent(Vk.LMENU, True)
    jmk_core(JmkEvent(Vk.A, False))
    assert jmk_core.next_handler.call_count == 2
    assert jmk_core.next_handler.call_args[0][0] == JmkEvent(Vk.LMENU, False)
    # tap
    jmk_core(JmkEvent(Vk.A, True))
    jmk_core(JmkEvent(Vk.A, False))
    assert jmk_core.next_handler.call_count == 4
    print(jmk_core.next_handler.call_args_list[-2])
    assert jmk_core.next_handler.call_args_list[-2][0][0] == JmkEvent(Vk.A, True)
    assert jmk_core.next_handler.call_args_list[-1][0][0] == JmkEvent(Vk.A, False)
    # pass through
    jmk_core(JmkEvent(Vk.B, True))
    jmk_core(JmkEvent(Vk.B, False))
    assert jmk_core.next_handler.call_count == 6


def test_jmk_core_layers(mocker):
    jmk_core = JmkCore(
        next_handler=mocker.Mock(),
        layers=[
            {Vk.G: JmkTapHold(hold=1, tap=Vk.G, term=0.3)},  # layer 0
            {Vk.H: JmkKey(Vk.LEFT)},  # layer 1
        ],
    )
    # simple toggle
    jmk_core(JmkEvent(Vk.G, True))
    time.sleep(0.4)
    assert jmk_core.next_handler.call_count == 0
    jmk_core(JmkEvent(Vk.H, True))
    assert jmk_core.next_handler.call_count == 1
    assert jmk_core.next_handler.call_args[0][0] == JmkEvent(Vk.LEFT, True)
    # toggle while holding
    jmk_core(JmkEvent(Vk.G, True))
    time.sleep(0.4)
    jmk_core(JmkEvent(Vk.H, True))
    jmk_core(JmkEvent(Vk.G, False))
    jmk_core(JmkEvent(Vk.H, False))
    assert jmk_core.next_handler.call_count == 3
    assert jmk_core.next_handler.call_args_list[-2][0][0] == JmkEvent(Vk.LEFT, True)
    assert jmk_core.next_handler.call_args_list[-1][0][0] == JmkEvent(Vk.LEFT, False)


def test_tap_tap_hold_passthrough(mocker):
    jmk_core = JmkCore(
        next_handler=mocker.Mock(),
        layers=[
            {
                Vk.A: JmkTapHold(
                    hold=Vk.LMENU,
                    tap=Vk.A,
                    quick_tap_term=0.2,
                ),
            }
        ],
    )
    jmk_core(JmkEvent(Vk.A, True))
    jmk_core(JmkEvent(Vk.A, False))
    time.sleep(0.1)
    assert jmk_core.next_handler.call_args_list[0][0][0] == JmkEvent(Vk.A, True)
    assert jmk_core.next_handler.call_args_list[1][0][0] == JmkEvent(Vk.A, False)
    jmk_core(JmkEvent(Vk.A, True))
    assert jmk_core.next_handler.call_args_list[2][0][0] == JmkEvent(Vk.A, True)
    jmk_core(JmkEvent(Vk.A, True))
    assert jmk_core.next_handler.call_args_list[3][0][0] == JmkEvent(Vk.A, True)
    jmk_core(JmkEvent(Vk.A, False))
    assert jmk_core.next_handler.call_args_list[4][0][0] == JmkEvent(Vk.A, False)



import enum
from ctypes import HRESULT, POINTER, pointer
from ctypes.wintypes import *
from typing import Optional

import comtypes
from comtypes import COMMETHOD, GUID, IUnknown


class DESKTOP_WALLPAPER_POSITION(enum.IntEnum):
    DWPOS_CENTER = 0
    DWPOS_TILE = 1
    DWPOS_STRETCH = 2
    DWPOS_FIT = 3
    DWPOS_FILL = 4
    DWPOS_SPAN = 5


# from https://stackoverflow.com/a/74203777/20763223
# noinspection PyPep8Naming
class IDesktopWallpaper(IUnknown):
    # Ref: https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nn-shobjidl_core-idesktopwallpaper

    # Search `IDesktopWallpaper` in `\HKEY_CLASSES_ROOT\Interface` to obtain the magic string
    _iid_ = GUID("{B92B56A9-8B55-4E14-9A89-0199BBB6F93B}")

    _methods_ = [
        COMMETHOD(
            [],
            HRESULT,
            "SetWallpaper",
            (["in"], LPCWSTR, "monitorID"),
            (["in"], LPCWSTR, "wallpaper"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "GetWallpaper",
            (["in"], LPCWSTR, "monitorID"),
            (["out"], POINTER(LPWSTR), "wallpaper"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "GetMonitorDevicePathAt",
            (["in"], UINT, "monitorIndex"),
            (["out"], POINTER(LPWSTR), "monitorID"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "GetMonitorDevicePathCount",
            (["out"], POINTER(UINT), "count"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "GetMonitorRECT",
            (["in"], LPCWSTR, "monitorID"),
            (["out"], POINTER(RECT), "displayRect"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "SetBackgroundColor",
            (["in"], COLORREF, "color"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "GetBackgroundColor",
            (["OUT"], POINTER(COLORREF), "color"),
        ),
    ]

    def SetWallpaper(self, monitor_path: str, wallpaper: str):
        self.__com_SetWallpaper(LPCWSTR(monitor_path), LPCWSTR(wallpaper))

    def GetWallpaper(self, monitorId: Optional[str] = None) -> str:
        """Retrieves wallpaper image path 

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nf-shobjidl_core-idesktopwallpaper-getwallpaper

        :param str monitor_path: optional, the string returned by GetMonitorDevicePathAt, return desktop wallpaper if omitted
        :returns: wallpaper image path
        :rtype: str
        """
        wallpaper = LPWSTR()
        self.__com_GetWallpaper(LPCWSTR(monitorId), pointer(wallpaper))
        return wallpaper.value

    def GetMonitorDevicePathAt(self, monitor_index: int) -> str:
        """Retrieves the unique ID of one of the system's monitors.

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nf-shobjidl_core-idesktopwallpaper-getmonitordevicepathat
        :param int monitor_index: monitor index
        :returns: monitor path
        :rtype: str
        """
        monitorId = LPWSTR()
        self.__com_GetMonitorDevicePathAt(UINT(monitor_index), pointer(monitorId))
        return monitorId.value

    def GetMonitorDevicePathCount(self) -> int:
        """Retrieves the number of monitors that are associated with the system.

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nf-shobjidl_core-idesktopwallpaper-getmonitordevicepathcount

        :returns: total number of available monitor (mirrored monitors excluded)
        """
        count = UINT()
        self.__com_GetMonitorDevicePathCount(pointer(count))
        return count.value

    def GetMonitorRECT(self, monitor_path: str) -> RECT:
        """Retrieves the display rectangle of the specified monitor.

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nf-shobjidl_core-idesktopwallpaper-getmonitorrect

        :param str monitor_path: monitor path returned by GetMonitorDevicePathAt
        :returns: display rectangle
        :rtype: RECT
        """
        rect = RECT()
        self.__com_GetMonitorRECT(LPCWSTR(monitor_path), pointer(rect))
        return rect

    def SetBackgroundColor(self, color: COLORREF):
        """Sets the background color
        
        :param COLORREF color:  background color
        """
        self.__com_SetBackgroundColor(color)

    def GetBackgroundColor(self) -> COLORREF:
        """Retrieves current background color
        
        :returns: current desktop background color
        :rtype: COLORREF
        """
        color = COLORREF()
        self.__com_GetBackgroundColor(pointer(color))
        return color


# Search `Desktop Wallpaper` in `\HKEY_CLASSES_ROOT\CLSID` to obtain the magic string
desktop_wallpaper: IDesktopWallpaper = comtypes.CoCreateInstance(
    GUID("{C2CF3110-460E-4fc1-B9D0-8A1C0C9CC4BD}"), interface=IDesktopWallpaper
)


def hexstr2colorref(hexstr: str) -> COLORREF:
    """Converts html color to Win32 COLORREF struct
    
    :param str hexstr: html color string
    :returns: Win32 color struct
    :rtype: COLORREF
    """
    color = int(hexstr[1:], 16)
    return COLORREF(color & 0xFF00 | (color >> 16 & 0xFF) | ((color & 0xFF) << 16))


def colorref2hexstr(color: COLORREF) -> str:
    """Converts Win32 COLORREF struct to html color

    :param COLORREF color: Win32 color struct
    :returns: html color string
    :rtype: str
    """
    color = int(color.value)
    return "#%x%x%x" % (color & 0xFF, color >> 8 & 0xFF, color >> 16 & 0xFF)


if __name__ == "__main__":
    # set_wallpaper(r"D:\Documents\wallpapers\IMG_20220806_151544.jpg")
    # set_wallpaper("")
    # print(get_wallpaper())
    # print(monitor0_path)
    # monitor0_path = desktop_wallpaper.GetMonitorDevicePathAt(0)
    # monitor0_rect = desktop_wallpaper.GetMonitorRECT(monitor0_path)
    # print(
    #     monitor0_path,
    #     monitor0_rect.left,
    #     monitor0_rect.top,
    #     monitor0_rect.right,
    #     monitor0_rect.bottom,
    # )

    hexstr = "#121314"
    colorref = hexstr2colorref(hexstr)
    print("hexstr", hexstr)
    print("colorref: %x" % colorref.value)
    print("backto hexstr: %s" % colorref2hexstr(colorref))



import json
import os


def open_fav_folder(root, fav_folder):
    """Opens the Chrome Favorites folder"""
    path = os.path.join(
        os.getenv("LOCALAPPDATA"),
        "Google",
        "Chrome",
        "User Data",
        "Default",
        "Bookmarks",
    )
    bookmarks = None
    with open(path, encoding="utf8") as f:
        bookmarks = json.load(f)

    # open folder recursively
    def open_folder(folder):
        for child in folder["children"]:
            if child.get("type") == "url":
                os.startfile(child["url"])
            else:
                open_folder(child)

    # locate the initial folder
    folder = bookmarks["roots"][root]
    for component in fav_folder.split("."):
        for child in folder["children"]:
            if child["name"] == component and child["type"] == "folder":
                folder = child
                break
    open_folder(folder)



import enum
import os
import random
import time
import typing
from ctypes import *
from ctypes.wintypes import *

from .vk import Vk

user32 = WinDLL("user32", use_last_error=True)

ULONG_PTR = LPARAM


class INPUTTYPE(enum.IntEnum):
    MOUSE = 0
    KEYBOARD = 1
    HARDWARE = 2


class KEYEVENTF(enum.IntFlag):
    EXTENDEDKEY = 0x0001
    KEYUP = 0x0002
    SCANCODE = 0x0008
    UNICODE = 0x0004


class KEYBDINPUT(Structure):
    _fields_ = (
        ("wVk", WORD),
        ("wScan", WORD),
        ("dwFlags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", ULONG_PTR),
    )


class MOUSEEVENTF(enum.IntFlag):
    ABSOLUTE = 0x8000
    HWHEEL = 0x01000
    MOVE = 0x0001
    MOVE_NOCOALESCE = 0x2000
    LEFTDOWN = 0x0002
    LEFTUP = 0x0004
    RIGHTDOWN = 0x0008
    RIGHTUP = 0x0010
    MIDDLEDOWN = 0x0020
    MIDDLEUP = 0x0040
    VIRTUALDESK = 0x4000
    WHEEL = 0x0800
    XDOWN = 0x0080
    XUP = 0x0100


class MOUSEINPUT(Structure):
    _fields_ = (
        ("dx", LONG),
        ("dy", LONG),
        ("mouseData", DWORD),
        ("dwFlags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", ULONG_PTR),
    )


class HARDWAREINPUT(Structure):
    _fields_ = (
        ("uMsg", LONG),
        ("wParamL", WORD),
        ("wParamH", WORD),
    )


class INPUTDATA(Union):
    _fields_ = (
        ("mi", MOUSEINPUT),
        ("ki", KEYBDINPUT),
        ("hi", HARDWAREINPUT),
    )


class INPUT(Structure):
    _anonymous_ = ["u"]
    _fields_ = (
        ("type", DWORD),
        ("u", INPUTDATA),
    )


random.seed(os.getpid())


def random_flags(length: int = 4, a: int = 0, b: int = 31) -> int:
    flags = set()
    while len(flags) < length:
        flags.add(1 << random.randint(a, b))
    return flags


def combine_flags(flags: typing.Iterable[int]) -> int:
    flag = 0
    for f in flags:
        flag |= f
    return flag


FLAGS = random_flags()
SYNTHESIZED_FLAG = combine_flags(FLAGS)


def send_input(*inputs: typing.List[INPUT], extra: int = 0):
    global SYNTHESIZED_FLAG
    """Synthesizes keystrokes, mouse motions, and button clicks.

    Usage:
    ```
    send_input(
        INPUT(
            type=INPUTTYPE.KEYBOARD,
            ki=KEYBDINPUT(wVk=VirtualKey.A),
        ),
        INPUT(
            type=INPUTTYPE.KEYBOARD,
            ki=KEYBDINPUT(wVk=VirtualKey.A, dwFlags=KEYEVENTF.KEYUP),
        ),
    )
    ```

    :param List[INPUT] inputs: list of keyboard/mouse inputs which
        to be sent to system

    """
    for item in inputs:
        if item is None:
            continue
        if item.type == INPUTTYPE.KEYBOARD:
            item.ki.dwExtraInfo = ULONG_PTR(extra | SYNTHESIZED_FLAG)
        elif item.type == INPUTTYPE.MOUSE:
            item.mi.dwExtraInfo = ULONG_PTR(extra | SYNTHESIZED_FLAG)
    length = len(inputs)
    array = INPUT * length
    if not user32.SendInput(length, array(*inputs), sizeof(INPUT)):
        raise WinError(get_last_error())


def is_synthesized(msg: typing.Union[KEYBDINPUT, MOUSEINPUT]) -> bool:
    """Check if keyboard/mouse event is sent by this module"""
    # the propability of conflict is 31 x 30 x 29 x 28 ...
    global FLAGS
    for flag in FLAGS:
        if not msg.dwExtraInfo & flag:
            return False
    return True


def set_synthesized_flag(flag: int):
    global SYNTHESIZED_FLAG
    SYNTHESIZED_FLAG = flag


def vk_to_input(vk: Vk, pressed: bool = None) -> typing.Optional[INPUT]:
    if vk < Vk.MS_BOUND or vk > Vk.KB_BOUND:
        dwFlags = 0
        mouseData = 0
        if vk == Vk.LBUTTON:
            if pressed:
                dwFlags = MOUSEEVENTF.LEFTDOWN
            else:
                dwFlags = MOUSEEVENTF.LEFTUP
        elif vk == Vk.RBUTTON:
            if pressed:
                dwFlags = MOUSEEVENTF.RIGHTDOWN
            else:
                dwFlags = MOUSEEVENTF.RIGHTUP
        elif vk == Vk.MBUTTON:
            if pressed:
                dwFlags = MOUSEEVENTF.MIDDLEDOWN
            else:
                dwFlags = MOUSEEVENTF.MIDDLEUP
        elif vk == Vk.XBUTTON1:
            mouseData = 0x0001
            if pressed:
                dwFlags = MOUSEEVENTF.XDOWN
            else:
                dwFlags = MOUSEEVENTF.XUP
        elif vk == Vk.XBUTTON2:
            mouseData = 0x0002
            if pressed:
                dwFlags = MOUSEEVENTF.XDOWN
            else:
                dwFlags = MOUSEEVENTF.XUP
        elif vk == Vk.WHEEL_UP:
            mouseData = 120
            dwFlags = MOUSEEVENTF.WHEEL
        elif vk == Vk.WHEEL_DOWN:
            mouseData = -120
            dwFlags = MOUSEEVENTF.WHEEL
        return INPUT(
            type=INPUTTYPE.MOUSE,
            mi=MOUSEINPUT(dwFlags=dwFlags, mouseData=mouseData),
        )
    else:
        dwFlags = 0
        if not pressed:
            dwFlags |= KEYEVENTF.KEYUP
        if vk >= vk.PRIOR and vk <= vk.HELP:
            dwFlags |= KEYEVENTF.EXTENDEDKEY
        return INPUT(
            type=INPUTTYPE.KEYBOARD,
            ki=KEYBDINPUT(wVk=vk, dwFlags=dwFlags),
        )


def reset_modifiers():
    send_input(
        vk_to_input(key, pressed=False)
        for key in [Vk.LSHIFT, Vk.RSHIFT, Vk.LCONTROL, Vk.RCONTROL, Vk.LMENU, Vk.RMENU]
    )


def send_combination(*comb: typing.Sequence[Vk]):
    # reset_modifiers()
    # press keys in combination in order
    for key in comb:
        send_input(vk_to_input(key, pressed=True))
    # release keys in combination in reverse order
    for key in reversed(comb):
        send_input(vk_to_input(key, pressed=False))


def send_text(text: str):
    """Send unicode text including emojis to active window. NOTE: do NOT use this in side
    a Hotkey with ALT modifier, it won't work and I don't know how to mitigate(sending ALT up didn't work).
    """
    b = text.encode("utf_16_le")
    for i in range(0, len(b), 2):
        code = b[i] | b[i + 1] << 8
        send_input(
            INPUT(
                type=INPUTTYPE.KEYBOARD,
                ki=KEYBDINPUT(dwFlags=KEYEVENTF.UNICODE, wScan=code),
            ),
            INPUT(
                type=INPUTTYPE.KEYBOARD,
                ki=KEYBDINPUT(dwFlags=KEYEVENTF.UNICODE | KEYEVENTF.KEYUP, wScan=code),
            ),
        )


if __name__ == "__main__":
    import time

    from .vk import Vk

    # print(hex(ord("😂")))
    # encodings = [
    #     "utf_32",
    #     "utf_32_be",
    #     "utf_32_le",
    #     "utf_16",
    #     "utf_16_be",
    #     "utf_16_le",
    #     "utf_7",
    #     "utf_8",
    #     "utf_8_sig",
    # ]
    # for e in encodings:
    #     b = "😄".encode(e)
    #     print(e.ljust(20), " ".join("{:02x}".format(x) for x in b))
    time.sleep(3)
    # send_text("hello😂您好")
    from datetime import datetime

    send_text(datetime.now().strftime("%Y-%m-%d"))

    # send_input(
    #     vk_to_input(Vk.WHEEL_UP),
    #     vk_to_input(Vk.WHEEL_DOWN),
    #     vk_to_input(Vk.XBUTTON2, pressed=True),
    #     vk_to_input(Vk.XBUTTON2, pressed=False),
    #     INPUT(
    #         type=INPUTTYPE.KEYBOARD,
    #         ki=KEYBDINPUT(wVk=Vk.LSHIFT),
    #     ),
    #     INPUT(
    #         type=INPUTTYPE.KEYBOARD,
    #         ki=KEYBDINPUT(wVk=Vk.INSERT, dwFlags=KEYEVENTF.EXTENDEDKEY),
    #     ),
    #     INPUT(
    #         type=INPUTTYPE.KEYBOARD,
    #         ki=KEYBDINPUT(
    #             wVk=Vk.INSERT, dwFlags=KEYEVENTF.KEYUP | KEYEVENTF.EXTENDEDKEY
    #         ),
    #     ),
    #     INPUT(
    #         type=INPUTTYPE.KEYBOARD,
    #         ki=KEYBDINPUT(wVk=Vk.LSHIFT, dwFlags=KEYEVENTF.KEYUP),
    #     ),
    # )



import enum
import typing


class Vk(enum.IntEnum):
    """Win32 virtual key code

    Ref: https://learn.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes

    """

    LBUTTON = 0x01  # Left mouse button
    RBUTTON = 0x02  # Right mouse button
    CANCEL = 0x03  # Control-break processing
    MBUTTON = 0x04  # Middle mouse button (three-button mouse)
    XBUTTON1 = 0x05  # X1 mouse button
    XBUTTON2 = 0x06  # X2 mouse button
    MS_BOUND = 0x07  # Maximum number of mouse buttons supported
    #  0x07  # Undefined
    BACK = 0x08  # BACKSPACE key
    TAB = 0x09  # TAB key
    #  0x0A-0B  # Reserved
    CLEAR = 0x0C  # CLEAR key
    RETURN = 0x0D  # ENTER key
    #  0x0E-0F  # Undefined
    WIN = 0x0E  # THIS UNDEFINED, hacked for shortcut
    SHIFT = 0x10  # SHIFT key
    CONTROL = 0x11  # CTRL key
    MENU = 0x12  # ALT key
    PAUSE = 0x13  # PAUSE key
    CAPITAL = 0x14  # CAPS LOCK key
    KANA = 0x15  # IME Kana mode
    HANGUEL = 0x15  # IME Hanguel mode (maintained for compatibility; use VK_HANGUL)
    HANGUL = 0x15  # IME Hangul mode
    IME_ON = 0x16  # IME On
    JUNJA = 0x17  # IME Junja mode
    FINAL = 0x18  # IME final mode
    HANJA = 0x19  # IME Hanja mode
    KANJI = 0x19  # IME Kanji mode
    IME_OFF = 0x1A  # IME Off
    ESCAPE = 0x1B  # ESC key
    CONVERT = 0x1C  # IME convert
    NONCONVERT = 0x1D  # IME nonconvert
    ACCEPT = 0x1E  # IME accept
    MODECHANGE = 0x1F  # IME mode change request
    SPACE = 0x20  # SPACEBAR
    PRIOR = 0x21  # PAGE UP key
    NEXT = 0x22  # PAGE DOWN key
    END = 0x23  # END key
    HOME = 0x24  # HOME key
    LEFT = 0x25  # LEFT ARROW key
    UP = 0x26  # UP ARROW key
    RIGHT = 0x27  # RIGHT ARROW key
    DOWN = 0x28  # DOWN ARROW key
    SELECT = 0x29  # SELECT key
    PRINT = 0x2A  # PRINT key
    EXECUTE = 0x2B  # EXECUTE key
    SNAPSHOT = 0x2C  # PRINT SCREEN key
    INSERT = 0x2D  # INS key
    DELETE = 0x2E  # DEL key
    HELP = 0x2F  # HELP key

    # Digits, the last 4 bits of the code represent the corresponding digit

    KEY_0 = 0x30  # '0' key
    KEY_1 = 0x31  # '1' key
    KEY_2 = 0x32  # '2' key
    KEY_3 = 0x33  # '3' key
    KEY_4 = 0x34  # '4' key
    KEY_5 = 0x35  # '5' key
    KEY_6 = 0x36  # '6' key
    KEY_7 = 0x37  # '7' key
    KEY_8 = 0x38  # '8' key
    KEY_9 = 0x39  # '9' key

    #  = 0x3A to 0x40 are undefined

    # The alphabet, the code corresponds to the capitalized letter in the ASCII code

    A = 0x41  # 'A' key
    B = 0x42  # 'B' key
    C = 0x43  # 'C' key
    D = 0x44  # 'D' key
    E = 0x45  # 'E' key
    F = 0x46  # 'F' key
    G = 0x47  # 'G' key
    H = 0x48  # 'H' key
    I = 0x49  # 'I' key
    J = 0x4A  # 'J' key
    K = 0x4B  # 'K' key
    L = 0x4C  # 'L' key
    M = 0x4D  # 'M' key
    N = 0x4E  # 'N' key
    O = 0x4F  # 'O' key
    P = 0x50  # 'P' key
    Q = 0x51  # 'Q' key
    R = 0x52  # 'R' key
    S = 0x53  # 'S' key
    T = 0x54  # 'T' key
    U = 0x55  # 'U' key
    V = 0x56  # 'V' key
    W = 0x57  # 'W' key
    X = 0x58  # 'X' key
    Y = 0x59  # 'Y' key
    Z = 0x5A  # 'Z' key

    LWIN = 0x5B  # Left Windows key (Natural keyboard)
    RWIN = 0x5C  # Right Windows key (Natural keyboard)
    APPS = 0x5D  # Applications key (Natural keyboard)
    # - = 0x5E  # Reserved
    SLEEP = 0x5F  # Computer Sleep key
    NUMPAD0 = 0x60  # Numeric keypad 0 key
    NUMPAD1 = 0x61  # Numeric keypad 1 key
    NUMPAD2 = 0x62  # Numeric keypad 2 key
    NUMPAD3 = 0x63  # Numeric keypad 3 key
    NUMPAD4 = 0x64  # Numeric keypad 4 key
    NUMPAD5 = 0x65  # Numeric keypad 5 key
    NUMPAD6 = 0x66  # Numeric keypad 6 key
    NUMPAD7 = 0x67  # Numeric keypad 7 key
    NUMPAD8 = 0x68  # Numeric keypad 8 key
    NUMPAD9 = 0x69  # Numeric keypad 9 key
    MULTIPLY = 0x6A  # Multiply key
    ADD = 0x6B  # Add key
    SEPARATOR = 0x6C  # Separator key
    SUBTRACT = 0x6D  # Subtract key
    DECIMAL = 0x6E  # Decimal key
    DIVIDE = 0x6F  # Divide key
    F1 = 0x70  # F1 key
    F2 = 0x71  # F2 key
    F3 = 0x72  # F3 key
    F4 = 0x73  # F4 key
    F5 = 0x74  # F5 key
    F6 = 0x75  # F6 key
    F7 = 0x76  # F7 key
    F8 = 0x77  # F8 key
    F9 = 0x78  # F9 key
    F10 = 0x79  # F10 key
    F11 = 0x7A  # F11 key
    F12 = 0x7B  # F12 key
    F13 = 0x7C  # F13 key
    F14 = 0x7D  # F14 key
    F15 = 0x7E  # F15 key
    F16 = 0x7F  # F16 key
    F17 = 0x80  # F17 key
    F18 = 0x81  # F18 key
    F19 = 0x82  # F19 key
    F20 = 0x83  # F20 key
    F21 = 0x84  # F21 key
    F22 = 0x85  # F22 key
    F23 = 0x86  # F23 key
    F24 = 0x87  # F24 key
    # - = 0x88-8F  # Unassigned
    NUMLOCK = 0x90  # NUM LOCK key
    SCROLL = 0x91  # SCROLL LOCK key
    # 0x92-96 = OEM specific
    # - = 0x97-9F  # Unassigned
    LSHIFT = 0xA0  # Left SHIFT key
    RSHIFT = 0xA1  # Right SHIFT key
    LCONTROL = 0xA2  # Left CONTROL key
    RCONTROL = 0xA3  # Right CONTROL key
    LMENU = 0xA4  # Left ALT key
    RMENU = 0xA5  # Right ALT key
    BROWSER_BACK = 0xA6  # Browser Back key
    BROWSER_FORWARD = 0xA7  # Browser Forward key
    BROWSER_REFRESH = 0xA8  # Browser Refresh key
    BROWSER_STOP = 0xA9  # Browser Stop key
    BROWSER_SEARCH = 0xAA  # Browser Search key
    BROWSER_FAVORITES = 0xAB  # Browser Favorites key
    BROWSER_HOME = 0xAC  # Browser Start and Home key
    VOLUME_MUTE = 0xAD  # Volume Mute key
    VOLUME_DOWN = 0xAE  # Volume Down key
    VOLUME_UP = 0xAF  # Volume Up key
    MEDIA_NEXT_TRACK = 0xB0  # Next Track key
    MEDIA_PREV_TRACK = 0xB1  # Previous Track key
    MEDIA_STOP = 0xB2  # Stop Media key
    MEDIA_PLAY_PAUSE = 0xB3  # Play/Pause Media key
    LAUNCH_MAIL = 0xB4  # Start Mail key
    LAUNCH_MEDIA_SELECT = 0xB5  # Select Media key
    LAUNCH_APP1 = 0xB6  # Start Application 1 key
    LAUNCH_APP2 = 0xB7  # Start Application 2 key
    # - = 0xB8-B9  # Reserved
    OEM_1 = 0xBA  # Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the ';:' key
    OEM_PLUS = 0xBB  # For any country/region, the '+' key
    OEM_COMMA = 0xBC  # For any country/region, the ',' key
    OEM_MINUS = 0xBD  # For any country/region, the '-' key
    OEM_PERIOD = 0xBE  # For any country/region, the '.' key
    OEM_2 = 0xBF  # Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the '/?' key
    OEM_3 = 0xC0  # Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the '`~' key
    # - = 0xC1-D7  # Reserved
    # - = 0xD8-DA  # Unassigned
    OEM_4 = 0xDB  # Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the '[{' key
    OEM_5 = 0xDC  # Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the '\|' key
    OEM_6 = 0xDD  # Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the ']}' key
    OEM_7 = 0xDE  # Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the 'single-quote/double-quote' key
    OEM_8 = 0xDF  # Used for miscellaneous characters; it can vary by keyboard.
    # - = 0xE0  # Reserved
    # 0xE1 = OEM specific
    OEM_102 = 0xE2  # The <> keys on the US standard keyboard, or the \\| key on the non-US 102-key keyboard
    # 0xE3-E4 = OEM specific
    PROCESSKEY = 0xE5  # IME PROCESS key
    # 0xE6 = OEM specific
    PACKET = 0xE7  # Used to pass Unicode characters as if they were keystrokes. The VK_PACKET key is the low word of a 32-bit Virtual Key value used for non-keyboard input methods. For more information, see Remark in KEYBDINPUT, SendInput, WM_KEYDOWN, and WM_KEYUP
    # - = 0xE8  # Unassigned
    # 0xE9-F5 = OEM specific
    ATTN = 0xF6  # Attn key
    CRSEL = 0xF7  # CrSel key
    EXSEL = 0xF8  # ExSel key
    EREOF = 0xF9  # Erase EOF key
    PLAY = 0xFA  # Play key
    ZOOM = 0xFB  # Zoom key
    NONAME = 0xFC  # Reserved
    PA1 = 0xFD  # PA1 key
    OEM_CLEAR = 0xFE  # Clear key
    KB_BOUND = 0xFF  # Clear key

    UNKNOWN = 0x00

    WHEEL_UP = 0x1000
    WHEEL_DOWN = 0x1001

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


VkAliases: typing.Dict[str, Vk] = {
    "LCTRL": Vk.LCONTROL,
    "LCTL": Vk.LCONTROL,
    "LALT": Vk.LMENU,
    "LSFT": Vk.LSHIFT,
    "LSUPER": Vk.LWIN,
    "RCTRL": Vk.RCONTROL,
    "RCTL": Vk.RCONTROL,
    "RALT": Vk.RMENU,
    "RSFT": Vk.RSHIFT,
    "RSUPER": Vk.RWIN,
    "CTRL": Vk.CONTROL,
    "MENU": Vk.MENU,
    "ALT": Vk.MENU,
    "SFT": Vk.SHIFT,
    "SUPER": Vk.WIN,
    "-": Vk.OEM_MINUS,
    "=": Vk.OEM_PLUS,
    ";": Vk.OEM_1,
    "/": Vk.OEM_2,
    "`": Vk.OEM_3,
    "[": Vk.OEM_4,
    "\\": Vk.OEM_5,
    "]": Vk.OEM_6,
    "'": Vk.OEM_7,
    ",": Vk.OEM_COMMA,
    ".": Vk.OEM_PERIOD,
}


class Vka:
    LCTRL = Vk.LCONTROL
    LCTL = Vk.LCONTROL
    LALT = Vk.LMENU
    LSFT = Vk.LSHIFT
    LSUPER = Vk.LWIN
    RCTRL = Vk.RCONTROL
    RCTL = Vk.RCONTROL
    RALT = Vk.RMENU
    RSFT = Vk.RSHIFT
    RSUPER = Vk.RWIN
    CTRL = Vk.CONTROL
    MENU = Vk.MENU
    ALT = Vk.MENU
    SFT = Vk.SHIFT
    SUPER = Vk.WIN
    DASH = Vk.OEM_MINUS
    EQUAL = Vk.OEM_PLUS
    SEMICOLON = Vk.OEM_1
    SLASH = Vk.OEM_2
    BACKQUOTE = Vk.OEM_3
    LBRACKET = Vk.OEM_4
    BACKSLASH = Vk.OEM_5
    RBRACKET = Vk.OEM_6
    SINGLEQUOTE = Vk.OEM_7
    COMMA = Vk.OEM_COMMA
    PERIOD = Vk.OEM_PERIOD


Modifers = {
    Vk.LCONTROL,
    Vk.RCONTROL,
    Vk.CONTROL,
    Vk.LSHIFT,
    Vk.RSHIFT,
    Vk.SHIFT,
    Vk.LMENU,
    Vk.RMENU,
    Vk.MENU,
    Vk.LWIN,
    Vk.RWIN,
    Vk.WIN,
}


def parse_key(key: str) -> Vk:
    key_name = key.strip().upper()
    # try alias
    key = VkAliases.get(key_name)
    # try name
    if key is None:
        if key_name not in Vk.__members__:
            raise Exception(f"invalid key: {key_name}")
        return Vk[key_name]
    return key


def parse_combination(combkeys: str) -> typing.Sequence[Vk]:
    """Converts combination in plain text ("Ctrl+s") to Sequence[Vk] ([Vk.CONTROL, Vk.S])"""
    parsed = []
    if not combkeys:
        return parsed
    for key_name in combkeys.split("+"):
        parsed.append(parse_key(key_name))
    return parsed


_key_expansions = {
    Vk.CONTROL: [Vk.LCONTROL, Vk.RCONTROL],
    Vk.MENU: [Vk.LMENU, Vk.RMENU],
    Vk.SHIFT: [Vk.LSHIFT, Vk.RSHIFT],
    Vk.WIN: [Vk.LWIN, Vk.RWIN],
}


def expand_combination(
    combkeys: typing.Sequence[Vk],
    index: typing.Optional[int] = 0,
) -> typing.Iterator[typing.Sequence[Vk]]:
    """Expand `Ctrl+s` to `LCtrl+s` and `RCtrl+s`, so on and so forth"""
    key = combkeys[index]
    expansions = _key_expansions.get(key)
    if expansions:
        is_last = index + 1 == len(combkeys)
        for mk in expansions:
            new_combkeys = combkeys[:index] + [mk]
            if is_last:
                yield new_combkeys
            else:
                yield from expand_combination(
                    new_combkeys + combkeys[index + 1 :], index + 1
                )
    else:
        yield combkeys



import os
from datetime import datetime, time
from typing import Optional

from jigsawwm.state import get_state_manager
from jigsawwm.w32.process import is_exe_running


def daily_once(name: str, day_start: Optional[time] = time(hour=8)):
    """Returns True if the given name has not been called today"""
    now = datetime.now().astimezone()
    if now.time() < day_start:
        return False
    today = now.date()
    state = get_state_manager()
    last_date = state.getdate("daily", name)
    state.setdate("daily", name, today)
    state.save()
    return last_date != today


def start_if_not_running(exe_path: str):
    """Returns True if the given name has not been called today"""
    if not is_exe_running(exe_path):
        os.startfile(exe_path)



"""

The ``tiler`` module is responsible for converting Layout to Physical Coordinates for arbitrary number of windows.

``Rect`` is a tuple with 4 elements (left/top/right/bottom) to describe a rectangle in pixels (integer)

``Tiler`` is a generator which generates Rects with specified Layout for given monitor work area and total number of windows

``LayoutTiler`` is similar to `Tiler` except the Layout was predefined

"""
from typing import Callable, Iterator, Tuple

from .layouts import Layout, dwindle, mono, plug_rect, widescreen_dwindle

# Rect holds physical coordinate for rectangle (left/top/right/bottom)
Rect = Tuple[int, int, int, int]

# Tiler generates physical Rects for specified total number of windows based on given Layout
Tiler = Callable[[Layout, Rect, int], Iterator[Rect]]

# LayoutTiler generates physical Rects for specified total number of windows
LayoutTiler = Callable[[Rect, int], Iterator[Rect]]


def direct_tiler(layout: Layout, work_area: Rect, total_windows: int) -> Iterator[Rect]:
    """The default Tiler which maps specified `Layout` to physical work area directly

    :param layout: a Layout generator
    :param work_area: the monitor work area (taskbar excluded)
    :param total_windows: total number of windows
    :rtype: Iterator[Rect]
    """
    rects = layout(total_windows)
    w = work_area[2] - work_area[0]
    h = work_area[3] - work_area[1]
    is_portrait = w < h
    if is_portrait:
        # rotate 90 degree if monitor in portrait mode
        rects = map(lambda r: (r[1], r[0], r[3], r[2]), rects)
    for float_rect in rects:
        yield tuple(int(f) for f in plug_rect(float_rect, work_area))


def obs_tiler(
    layout: Layout,
    work_area: Rect,
    total_windows: int,
    obs_width: int = 1920,
    obs_height: int = 1080,
) -> Iterator[Rect]:
    """Useful for OBS screen recording, it would put the 1st and 2nd window to the left
    and bottom, while all the other windows go to the top right corne with specified
    layout as the area to be recorded.

    .. code-block:: text

        +-----------+-----------+-----------+
        |           |           |           |
        |           |           |     4     |
        |           |     3     |           |
        |     1     |           +-----+-----+
        |    obs    |           |  5  |  6  |
        |           |-----------------------|
        |           |   2. script reader    |
        +-----------+-----------------------+

    :param layout: `Layout` for the top right corne (area to be recorded)
    :param work_area: the monitor work area (taskbar excluded)
    :param total_windows: total number of windows
    :param obs_width: width for the top right corne (area to be recorded) in pixels
    :param obs_height: height for the top right corne (area to be recorded) in pixels
    :rtype: Iterator[Rect]
    """
    wl, wt, wr, wb = work_area
    scr_width, scr_height = wr - wl, wb - wt
    # fallback to direct_tiler when work_area is smaller than obs reserved area
    if obs_width >= scr_width or obs_height >= scr_height:
        yield from direct_tiler(layout, work_area, total_windows)
        return
    if total_windows == 0:
        return
    fr = wr - obs_width
    # first window on the left
    yield wl, wt, fr, wb
    if total_windows == 1:
        return
    # second window on the bottom right
    yield fr, wt + obs_height, wr, wb
    if total_windows == 2:
        return
    obs_rect = (fr, wt, wr, wt + obs_height)
    yield from direct_tiler(layout, obs_rect, total_windows - 2)


def mono_layout_tiler(*args, **kwargs) -> Iterator[Rect]:
    """The dwindle layout tiler"""
    return direct_tiler(mono, *args, **kwargs)


def dwindle_layout_tiler(*args, **kwargs) -> Iterator[Rect]:
    """The dwindle layout tiler"""
    return direct_tiler(dwindle, *args, **kwargs)


def widescreen_dwindle_layout_tiler(*args, **kwargs) -> Iterator[Rect]:
    """The wide-screen dwindle layout tiler"""
    return direct_tiler(widescreen_dwindle, *args, **kwargs)


def obs_dwindle_layout_tiler(*args, **kwargs) -> Iterator[Rect]:
    """The obs dwindle layout tiler"""
    return obs_tiler(dwindle, *args, **kwargs)


if __name__ == "__main__":
    print("direct dwindle")
    for n in range(1, 5):
        print(
            list(
                direct_tiler(
                    layout=dwindle,
                    work_area=(10, 10, 3450, 1450),
                    total_windows=n,
                )
            )
        )
    print("obs dwindle")
    for n in range(1, 5):
        print(
            list(
                obs_tiler(
                    layout=dwindle,
                    work_area=(10, 10, 3450, 1450),
                    total_windows=n,
                )
            )
        )



import os
from ctypes import *
from ctypes.wintypes import *
from typing import List

kernel32 = WinDLL("kernel32", use_last_error=True)
advapi32 = WinDLL("advapi32", use_last_error=True)
psapi = WinDLL("psapi", use_last_error=True)

TOKEN_QUERY = DWORD(8)


def open_process_for_limited_query(pid: int) -> HANDLE:
    """Opens an existing local process object with permission to query limited information

    Ref: https://learn.microsoft.com/en-us/windows/win32/procthread/process-security-and-access-rights

    :param int pid: process id
    :return:  process handle
    :rtype: HANDLE
    """
    PROCESS_QUERY_LIMITED_INFORMATION = DWORD(0x1000)
    hprc = kernel32.OpenProcess(PROCESS_QUERY_LIMITED_INFORMATION, False, pid)
    if not hprc:
        raise WinError(get_last_error())
    return hprc


def is_elevated(pid: int) -> bool:
    """Check if specified process is elevated (run in Administrator Role)

    :param int pid: process id
    :return: `True` if elevated, `False` otherwise
    :rtype: bool
    """
    hprc = None
    try:
        hprc = open_process_for_limited_query(pid)
    except:
        return True
    htoken = PHANDLE()
    if not windll.advapi32.OpenProcessToken(hprc, TOKEN_QUERY, byref(htoken)):
        windll.kernel32.CloseHandle(hprc)
        return
    TOKEN_ELEVATION = INT(20)
    is_elevated = BOOL()
    returned_length = DWORD()
    if not advapi32.GetTokenInformation(
        htoken,
        TOKEN_ELEVATION,
        byref(is_elevated),
        4,
        byref(returned_length),
    ):
        raise WinError(get_last_error())
    kernel32.CloseHandle(hprc)
    kernel32.CloseHandle(htoken)
    return bool(is_elevated.value)


def get_exepath(pid: int) -> str:
    """Retrieves the full name of the executable image for the specified process.

    :param int pid: process id
    :return: the full path of the executable
    :rtype: str
    """
    if not pid:
        return
    hprc = open_process_for_limited_query(pid)
    buff = create_unicode_buffer(512)
    size = DWORD(sizeof(buff))
    if not kernel32.QueryFullProcessImageNameW(hprc, 0, buff, pointer(size)):
        kernel32.CloseHandle(hprc)
        raise WinError(get_last_error())
    kernel32.CloseHandle(hprc)
    return str(buff.value)


def get_all_processes(total: int = 1024) -> List[DWORD]:
    """Retrieves the process identifiers of all running processes.

    :param in total: the number of processes to retrieve

    :return: list of process identifiers
    :rtype: List[DWORD]
    """
    buff = (DWORD * total)()
    size = DWORD(sizeof(buff))
    if not psapi.EnumProcesses(byref(buff), size, pointer(size)):
        raise WinError(get_last_error())
    return list(buff[: size.value // sizeof(DWORD)])


def is_exe_running(exe: str, nameonly: bool = False) -> bool:
    """Check if specified executable is running

    :param str exe: executable name
    :param bool nameonly: if `True`, only check the executable name, otherwise check the full path
    :return: `True` if running, `False` otherwise
    :rtype: bool
    """
    exe = exe.lower()
    if nameonly:
        exe = os.path.basename(exe)
    for pid in get_all_processes():
        try:
            ppath = get_exepath(pid).lower()
            if nameonly:
                ppath = os.path.basename(ppath)
            if exe == ppath:
                return True
        except:
            pass
    return False


if __name__ == "__main__":
    import sys

    print(is_exe_running(sys.argv[1], bool(sys.argv[2])))



import re
from ctypes.wintypes import DWORD, HWND, LONG
from queue import SimpleQueue
from typing import List, Set, Union

from jigsawwm.w32 import hook
from jigsawwm.w32.sendinput import is_synthesized, send_input, vk_to_input
from jigsawwm.w32.window import Window

from .core import *

q = SimpleQueue()


class SystemInput:
    """A handler that handles system input events.

    :param next_handler: the next handler in the chain, normally a JmkCore instance
    :param bypass_exe: a list of regular expression patterns that matches the exe path
        some applications (e.g. Windows 10's touch keyboard, emoji input) will not
        work properly, so we need to bypass them
    """

    hook_handles: List[hook.HHOOK] = None
    next_handler: JmkHandler
    focused_window: Window = None
    disabled: bool = False
    bypass_exe: List[re.Pattern] = None
    pressed_key: Set[Vk] = set()

    def __init__(self, next_handler: JmkHandler, bypass_exe: List[re.Pattern] = None):
        """Initialize a system input handler"""
        self.next_handler = next_handler
        self.bypass_exe = bypass_exe or [
            re.compile(r"^C:\\Windows\\.*\\TextInputHost.exe$", re.IGNORECASE)
        ]
        self.pressed_key = set()

    def install(self):
        self.hook_handles = [
            hook.hook_keyboard(self.input_event_handler),
            hook.hook_mouse(self.input_event_handler),
            hook.hook_winevent(
                hook.WinEvent.EVENT_OBJECT_FOCUS,
                hook.WinEvent.EVENT_OBJECT_FOCUS,
                self.winevent,
            ),
        ]

    def uninstall(self):
        for hook_handle in self.hook_handles:
            hook.unhook(hook_handle)

    def winevent(
        self,
        event: hook.WinEvent,
        hwnd: HWND,
        id_obj: LONG,
        id_chd: LONG,
        id_evt_thread: DWORD,
        time: DWORD,
    ):
        if self.focused_window is None or self.focused_window.handle != hwnd:
            self.focused_window = Window(hwnd)
            if self.focused_window.is_evelated:
                logger.debug("focused window is elevated, disable jmk")
                self.disabled = True
                return
            if self.bypass_exe:
                for pattern in self.bypass_exe:
                    if pattern.match(self.focused_window.exe):
                        logger.debug("focused window is in bypass list, disable jmk")
                        self.disabled = True
                        return
            logger.debug("focused window is not in bypass list, enable jmk")
            self.disabled = False

    def input_event_handler(
        self,
        msgid: Union[hook.KBDLLHOOKMSGID, hook.MSLLHOOKMSGID],
        msg: Union[hook.KBDLLHOOKDATA, hook.MSLLHOOKDATA],
    ) -> bool:
        """Handles keyboard events and call callback if the combination
        had been registered
        """
        if is_synthesized(msg):
            return False
        # convert keyboard/mouse event to a unified virtual key representation
        vkey, pressed = None, None
        if isinstance(msgid, hook.KBDLLHOOKMSGID):
            vkey = Vk(msg.vkCode)
            if vkey == Vk.PACKET:
                return False
            if (
                msgid == hook.KBDLLHOOKMSGID.WM_KEYDOWN
                or msgid == hook.KBDLLHOOKMSGID.WM_SYSKEYDOWN
            ):
                pressed = True
            elif (
                msgid == hook.KBDLLHOOKMSGID.WM_KEYUP
                or msgid == hook.KBDLLHOOKMSGID.WM_SYSKEYUP
            ):
                pressed = False
        elif isinstance(msgid, hook.MSLLHOOKMSGID):
            if msgid == hook.MSLLHOOKMSGID.WM_LBUTTONDOWN:
                vkey = Vk.LBUTTON
                pressed = True
            elif msgid == hook.MSLLHOOKMSGID.WM_LBUTTONUP:
                vkey = Vk.LBUTTON
                pressed = False
            elif msgid == hook.MSLLHOOKMSGID.WM_RBUTTONDOWN:
                vkey = Vk.RBUTTON
                pressed = True
            elif msgid == hook.MSLLHOOKMSGID.WM_RBUTTONUP:
                vkey = Vk.RBUTTON
                pressed = False
            elif msgid == hook.MSLLHOOKMSGID.WM_MBUTTONDOWN:
                vkey = Vk.MBUTTON
                pressed = True
            elif msgid == hook.MSLLHOOKMSGID.WM_MBUTTONUP:
                vkey = Vk.MBUTTON
                pressed = False
            elif msgid == hook.MSLLHOOKMSGID.WM_XBUTTONDOWN:
                vkey = Vk.XBUTTON1 if msg.hiword() == 1 else Vk.XBUTTON2
                pressed = True
            elif msgid == hook.MSLLHOOKMSGID.WM_XBUTTONUP:
                vkey = Vk.XBUTTON1 if msg.hiword() == 1 else Vk.XBUTTON2
                pressed = False
            elif msgid == hook.MSLLHOOKMSGID.WM_MOUSEWHEEL:
                delta = msg.get_wheel_delta()
                if delta > 0:
                    vkey = Vk.WHEEL_UP
                else:
                    vkey = Vk.WHEEL_DOWN
                pressed = False
        # skip events that out of our interest
        if vkey is None or pressed is None:
            return

        # bypass events when disabled unless it's a keyup event of a pressed key
        if self.disabled and (
            pressed or (not pressed and vkey not in self.pressed_key)
        ):
            return
        if pressed:
            self.pressed_key.add(vkey)
        elif vkey in self.pressed_key:
            self.pressed_key.remove(vkey)
        evt = JmkEvent(vkey, pressed, system=True, extra=msg.dwExtraInfo)
        logger.debug("sys >>> %s", evt)
        swallow = self.next_handler(evt)
        return swallow


class SystemOutput(JmkHandler):
    """A system output handler that send input to system

    :param always_swallow: whether to always swallow the event, should be `True`
        when use as a Mouse/Keyboard transformer to keep the input order as expect,
        or `False` if you just need to register a system hotkey. default to True
    """

    always_swallow: bool

    def __init__(self, always_swallow: bool = True):
        self.always_swallow = always_swallow

    def __call__(self, evt: JmkEvent) -> bool:
        if evt.system and not self.always_swallow:
            logger.debug("nil <<< %s", evt)
            return False
        logger.debug("sys <<< %s", evt)
        q.put(evt)
        return True


def consume_queue():
    while True:
        evt = q.get()
        send_input(vk_to_input(evt.vk, evt.pressed), extra=evt.extra)


executor.submit(consume_queue)



import logging
import time
import traceback
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field

from jigsawwm.w32.sendinput import send_combination
from jigsawwm.w32.vk import *

logger = logging.getLogger(__name__)
# for executing callback function
executor = ThreadPoolExecutor(max_workers=100)


def execute(func, *args, **kwargs):
    """Execute a function in a thread pool and print the exception if any

    :param func: the function to execute
    :param args: the arguments to pass to the function
    :param kwargs: the keyword arguments to pass to the function
    """

    def wrapped():
        logger.debug("[JmkCore] executing %s", func)
        try:
            func()
        except Exception as e:
            traceback.print_exception(e)

    return executor.submit(func, *args, **kwargs)


@dataclass
class JmkEvent:
    """A jmk event that contains the key/button, pressed state, system state(does it came from the OS) and extra data"""

    vk: Vk
    pressed: bool
    system: bool = False
    extra: int = 0
    time: float = field(default_factory=time.time)

    __repr__ = (
        lambda self: f"JmkEvent({self.vk.name}, {'down' if self.pressed else 'up'}, {'sys' if self.system else 'sim'}, {self.extra})"
    )


JmkHandler = typing.Callable[[JmkEvent], bool]


class JmkLayerKey(JmkHandler):
    """A key handler that can be used in a layer"""

    state: "JmkCore" = None
    layer: int = None
    vk: Vk = None

    __repr__ = (
        lambda self: f"{self.__class__.__name__}(layer={self.layer}, vk={self.vk.name})"
    )

    def other_key(self, evt: JmkEvent):
        pass


class JmkKey(JmkLayerKey):
    """The basic key handler in a layer

    :param kf: can be a Vk(map key to another key), a string, a list of Vk(map key to a combination) or a callable (a function to execute)
    :param swallow: whether to swallow the event, normally True
    """

    func: callable = None
    key: Vk = None
    swallow: bool

    def __init__(
        self,
        kf: typing.Union[Vk, str, typing.List[Vk], typing.Callable],
        swallow: bool = True,
    ):
        if isinstance(kf, str):
            kf = parse_combination(kf)
        if isinstance(kf, list):
            kf = lambda: send_combination(*kf)
        if callable(kf):
            self.func = kf
        else:
            self.key = kf
        self.swallow = swallow

    def __call__(self, evt: JmkEvent) -> bool:
        if self.func and not evt.pressed:
            logger.debug("%s <<< %s", "nil" if self.swallow else "sys", self)
            execute(self.func)
            return self.swallow
        elif self.key:
            logger.debug("%s >>>", self)
            return self.state.next_handler(JmkEvent(self.key, evt.pressed))
        return True


class JmkTapHold(JmkLayerKey):
    """A advanced key handler that can be used in a layer

    :param tap: map the key to another key when tapped
    :param hold: map the key to another key when hold
    :param on_hold_down: a function to execute when hold down
    :param on_hold_up: a function to execute when hold up
    :param on_tap: a function to execute when tapped
    :param term: the term to determine whether it is a hold
    :param quick_tap_term: tap a key and then hold it down within this term will enter quick tap mode,
                        when activated, the tap key will be sent as long as the key is hold. It is useful for
                        dual function keys like using key A as the Alt key and you want to input a bunch of A
    """

    tap: typing.Optional[Vk] = None
    hold: typing.Optional[typing.Union[Vk, str]] = None
    on_hold_down: typing.Optional[typing.Callable] = None
    on_hold_up: typing.Optional[typing.Callable] = None
    on_tap: typing.Optional[typing.Callable] = None
    term: float = 0.2
    quick_tap_term: float
    last_tapped_at: float = 0
    quick_tap: bool = False
    other_pressed_keys: typing.Set[Vk]
    resend: typing.List[JmkEvent]
    pressed: int = 0
    held: bool = False

    def __init__(
        self,
        tap: Vk = None,
        hold: Vk = None,
        on_hold_down: callable = None,
        on_hold_up: callable = None,
        on_tap: callable = None,
        term: float = 0.2,
        quick_tap_term: float = 0.12,
    ):
        self.tap = tap
        self.hold = hold
        self.on_hold_down = on_hold_down
        self.on_hold_up = on_hold_up
        self.on_tap = on_tap
        self.term = term
        self.quick_tap_term = quick_tap_term
        self.other_pressed_keys = set()
        self.resend = []
        self.pressed = 0
        self.held = False

    def check_hold(self):
        if time.time() - self.pressed > self.term:
            self.hold_down()

    def hold_down(self):
        if self.held:
            # hold_down might be triggered multiple places
            return
        self.held = True
        logger.debug("%s hold down", self)
        if self.on_hold_down:
            logger.debug("%s on_hold_down", self)
            execute(self.on_hold_down)
        if self.hold:
            if isinstance(self.hold, Vk):
                evt = JmkEvent(self.hold, True)
                logger.debug("%s %s down >>>", self, evt)
                self.state.next_handler(evt)
            else:
                self.state.activate_layer(self.hold)
        self.flush_resend()

    def hold_up(self):
        self.pressed = 0
        self.held = False
        self.other_pressed_keys.clear()
        logger.debug("%s hold up", self)
        if self.on_hold_up:
            logger.debug("%s on_hold_up", self)
            execute(self.on_hold_up)
        if self.hold:
            if isinstance(self.hold, Vk):
                evt = JmkEvent(self.hold, False)
                logger.debug("%s %s up >>>", self, evt)
                self.state.next_handler(evt)
            else:
                self.state.deactivate_layer(self.hold)

    def tap_down_up(self):
        self.pressed = 0
        self.held = False
        logger.debug("%s tapped", self)
        if self.tap:
            evt_down = JmkEvent(self.tap, True)
            logger.debug("%s %s >>>", self, evt_down)
            self.state.next_handler(evt_down)
            evt_up = JmkEvent(self.tap, False)
            logger.debug("%s %s >>>", self, evt_up)
            self.state.next_handler(evt_up)
        if self.on_tap:
            logger.debug("%s on_tap", self)
            execute(self.on_tap)
        self.last_tapped_at = time.time()
        self.flush_resend()

    def other_key(self, evt: JmkEvent) -> bool:
        # intercept timing: after key down, before hold/tap determined
        if not self.pressed or self.held:
            return False
        logger.debug("%s queue other key %s >>>", self, evt)
        self.resend.append(evt)
        if evt.pressed:
            self.other_pressed_keys.add(evt.vk)
        elif evt.vk in self.other_pressed_keys:
            # there was a key tapping, we shal get into the holding mode immediately
            self.other_pressed_keys.remove(evt.vk)
            self.hold_down()
        # wheel up/down doesn't have a key down event
        if evt.vk in (
            Vk.WHEEL_UP,
            Vk.WHEEL_DOWN,
        ):
            self.hold_down()
        # or timeout
        self.check_hold()
        # delay the key until we know if it's a tap or hold
        return True

    def flush_resend(self):
        if self.resend:
            for evt in self.resend:
                logger.debug("%s resend %s >>>", self, evt)
                self.state(evt)
        self.resend.clear()

    def __call__(self, evt: JmkEvent) -> bool:
        # quick tap check
        if (
            evt.pressed
            and self.last_tapped_at
            and time.time() - self.last_tapped_at < self.quick_tap_term
        ):
            self.quick_tap = True
        if self.quick_tap:
            if not evt.pressed:
                self.last_tapped_at = 0
                self.quick_tap = False
            evt = JmkEvent(self.tap or evt.vk, evt.pressed)
            logger.debug("%s quick tap %s >>>", self, evt)
            return self.state.next_handler(evt)
        # tap hold
        if evt.pressed:
            if not self.pressed:
                # initial state
                self.pressed = evt.time
            else:
                self.check_hold()
        else:
            # reset state
            self.check_hold()
            if self.held:
                self.hold_up()
            else:
                self.tap_down_up()
        return True


JmkLayer = typing.Dict[Vk, JmkHandler]


class JmkCore(JmkHandler):
    """JmkCore is the core of JmkHandler, it handles the key events and dispatches them to the registered handlers.

    :param next_handler: the next handler to dispatch the event to
    :param layers: the layers
    """

    next_handler: JmkHandler
    layers: typing.List[JmkLayer]
    active_layers: typing.Set[int]
    routes: JmkLayer

    def __init__(
        self,
        next_handler: JmkHandler,
        layers: typing.List[JmkLayer] = None,
    ):
        if len(layers) < 1:
            raise ValueError("layers must have at least one layer")
        self.layers = [{}]
        for index, layer in enumerate(layers):
            if not isinstance(layer, dict):
                raise TypeError("layer must be a dict")
            for vk, handler in layer.items():
                self.register(vk, handler, index)
        self.next_handler = next_handler
        self.active_layers = {0}
        self.routes = {}

    def register(self, vk: Vk, handler: JmkLayerKey, layer: int = 0):
        if not isinstance(vk, Vk):
            raise TypeError("layer key must be a Vk")
        if not isinstance(handler, JmkLayerKey):
            raise TypeError("layer value must be a JmkLayerKey")
        if handler.layer is not None:
            raise ValueError("layer value must not have layer index")
        handler.state = self
        handler.layer = layer
        handler.vk = vk
        while len(self.layers) <= layer:
            self.layers.append({})
        self.layers[layer][vk] = handler

    def check_index(self, index: int):
        if index < 1 or index >= len(self.layers):
            err = IndexError(f"layer index {index} out of range")
            logger.error(err)
            raise err

    def activate_layer(self, index: int):
        # self.check_index(index)
        logger.debug(f"activating layer {index}")
        self.active_layers.add(index)

    def deactivate_layer(self, index: int):
        # self.check_index(index)
        logger.debug(f"deactivating layer {index}")
        self.active_layers.remove(index)

    def find_route(self, vk: Vk) -> typing.Optional[JmkLayerKey]:
        i = len(self.layers) - 1
        while i >= 0:
            if i in self.active_layers and vk in self.layers[i]:
                return self.layers[i][vk]
            i -= 1

    def __call__(self, evt: JmkEvent) -> bool:
        # route is to handle situation that a key is still held down after layer turned off
        route = None
        for vk, rt in self.routes.items():
            if vk == evt.vk:
                route = rt
            elif rt.other_key(evt):
                # key is intercepted by other key, most likely a TapHold
                return True
        if route and not evt.pressed:
            self.routes.pop(evt.vk)
        elif not route:
            route = self.find_route(evt.vk)
            if route and evt.pressed:
                self.routes[evt.vk] = route
        if route:
            logger.debug("routing %s to %s", evt, route)
            return route(evt)
        return self.next_handler(evt)



"""

The ``layout`` module operates in Relative Coordinate, to that end, it defines 2 basic types:

``FloatRect`` is a tuple with 4 elements (left/top/right/bottom) to describe a rectangle in ratio form (0.0~1.0)

``Layout`` is a generator which generates FloatRects for given total number of windows

"""

from functools import partial
from typing import Callable, Iterator, Tuple, Union

# FloatRect holds relative coordinate for rectangle (left/top/right/bottom)
FloatRect = Tuple[float, float, float, float]

# Layout accepts an integer (total number of windows) and return a FloatRect generator
Layout = Callable[[int], Iterator[FloatRect]]


def mono(n: int) -> Iterator[FloatRect]:
    """The mono Layout

    .. code-block:: text

        +-----------+
        |           |
        |           |
        |           |
        |     1     |
        |           |
        |           |
        |           |
        +-----------+

    :param n: total number of windows
    :rtype: Iterator[FloatRect]
    """
    for i in range(n):
        yield 0.0, 0.0, 1.0, 1.0


def dwindle(n: int) -> Iterator[FloatRect]:
    """The dwindle Layout

    .. code-block:: text

        +-----------+-----------+
        |           |           |
        |           |     2     |
        |           |           |
        |     1     +-----+-----+
        |           |     |  4  |
        |           |  3  +--+--+
        |           |     | 5|-.|
        +-----------+-----+-----+

    :param n: total number of windows
    :rtype: Iterator[FloatRect]
    """
    l, t, r, b = 0.0, 0.0, 1.0, 1.0
    last_index = n - 1
    for i in range(n):
        # last window would occupy the whole area
        if i == last_index:
            yield l, t, r, b
        # or it should leave out half space for the other windows
        elif i % 2 == 0:
            nl = r - (r - l) / 2
            yield l, t, nl, b
            l = nl
        else:
            nb = b - (b - t) / 2
            yield l, t, r, nb
            t = nb


def widescreen_dwindle(n: int, master_ratio: float = 0.4) -> Iterator[FloatRect]:
    """A wide-screen friendly dwindle Layout

    .. code-block:: text

        +-----------+-----------+-----------+
        |           |           |           |
        |           |           |     3     |
        |           |           |           |
        |     1     |     2     +-----+-----+
        |           |           |     |  5  |
        |           |           |  4  +--+--+
        |           |           |     | 6|-.|
        +-----------+-----------+-----+-----+

    :param n: total number of windows
    :rtype: Iterator[FloatRect]
    """
    if n == 0:
        return
    #    wide_dwindle
    if n == 1:
        yield 0.0, 0.0, 1.0, 1.0
        return
    # master window on the left
    yield 0.0, 0.0, master_ratio, 1
    # other windows on the right with dwindle layout, just map the coordinate and we are good
    yield from map(
        partial(plug_rect, target=(master_ratio, 0.0, 1.0, 1.0)), dwindle(n - 1)
    )


Number = Union[int, float]
NumberRect = Tuple[Number, Number, Number, Number]


def plug_rect(source: FloatRect, target: NumberRect) -> NumberRect:
    """Plug the source rect into the target rect and compute the new dimensions,
    you may plug a Relative Rect into a Physical Rect, but not the other way around.

    :param source: the FloatRect to be moved
    :param target: the container, either a FloatRect or Rect(physical pixels)
    :returns: Rect or FloatRect depends on the type of the target
    :rtype: NumberRect
    """

    sl, st, sr, sb = source
    tl, tt, tr, tb = target
    tw = tr - tl
    th = tb - tt
    return (
        tl + sl * tw,  # left = target left + scaled source left
        tt + st * th,  # top = target top + scaled source top
        tl + sr * tw,  # right
        tt + sb * th,  # bottom
    )


if __name__ == "__main__":
    print("dwindle")
    for n in range(1, 5):
        print(list(dwindle(n)))
    print()
    print("widescreen_dwindle")
    for n in range(1, 5):
        print(list(widescreen_dwindle(n)))



from typing import Callable, List, Tuple

from .core import *
from .hotkey import *
from .sysinout import *


def create_jmk(
    layers: List[JmkLayer],
    hotkeys: List[Tuple[JmkHotkeyComb, Callable]],
    always_swallow: bool = True,
):
    # from tail to head
    sysout = SystemOutput(always_swallow)
    hks = JmkHotkeys(sysout, hotkeys)
    jmk = JmkCore(hks, layers)
    sysin = SystemInput(jmk)
    return sysin, jmk, hks, sysout



import enum
from ctypes import WINFUNCTYPE
from ctypes.wintypes import *

HWINEVENTHOOK = HANDLE

WINEVENTHOOKPROC = WINFUNCTYPE(
    HWINEVENTHOOK, HANDLE, DWORD, HWND, LONG, LONG, DWORD, DWORD
)


class WinEvent(enum.IntEnum):
    # The range of WinEvent constant values specified by the Accessibility Interoperability Alliance (AIA) for use across the industry. For more information, see Allocation of WinEvent IDs.
    EVENT_AIA_START = 0xA000
    EVENT_AIA_END = 0xAFFF

    # The lowest and highest possible event values.
    EVENT_MIN = 0x00000001
    EVENT_MAX = 0x7FFFFFFF

    EVENT_OBJECT_ACCELERATORCHANGE = 0x8012
    # Sent when a window is cloaked
    EVENT_OBJECT_CLOAKED = 0x8017
    EVENT_OBJECT_CONTENTSCROLLED = 0x8015
    EVENT_OBJECT_CREATE = 0x8000
    EVENT_OBJECT_DEFACTIONCHANGE = 0x8011
    EVENT_OBJECT_DESCRIPTIONCHANGE = 0x800D
    EVENT_OBJECT_DESTROY = 0x8001
    EVENT_OBJECT_DRAGSTART = 0x8021
    EVENT_OBJECT_DRAGCANCEL = 0x8022
    EVENT_OBJECT_DRAGCOMPLETE = 0x8023
    EVENT_OBJECT_DRAGENTER = 0x8024
    EVENT_OBJECT_DRAGLEAVE = 0x8025
    EVENT_OBJECT_DRAGDROPPED = 0x8026
    EVENT_OBJECT_END = 0x80FF
    EVENT_OBJECT_FOCUS = 0x8005
    EVENT_OBJECT_HELPCHANGE = 0x8010
    # An object is hidden.
    EVENT_OBJECT_HIDE = 0x8003
    EVENT_OBJECT_HOSTEDOBJECTSINVALIDATED = 0x8020
    EVENT_OBJECT_IME_HIDE = 0x8028
    EVENT_OBJECT_IME_SHOW = 0x8027
    EVENT_OBJECT_IME_CHANGE = 0x8029
    EVENT_OBJECT_INVOKED = 0x8013
    EVENT_OBJECT_LIVEREGIONCHANGED = 0x8019
    EVENT_OBJECT_LOCATIONCHANGE = 0x800B
    EVENT_OBJECT_NAMECHANGE = 0x800C
    EVENT_OBJECT_PARENTCHANGE = 0x800F
    EVENT_OBJECT_REORDER = 0x8004
    EVENT_OBJECT_SELECTION = 0x8006
    EVENT_OBJECT_SELECTIONADD = 0x8007
    EVENT_OBJECT_SELECTIONREMOVE = 0x8008
    EVENT_OBJECT_SELECTIONWITHIN = 0x8009
    # A hidden object is shown
    EVENT_OBJECT_SHOW = 0x8002
    EVENT_OBJECT_STATECHANGE = 0x800A
    EVENT_OBJECT_TEXTEDIT_CONVERSIONTARGETCHANGED = 0x8030
    EVENT_OBJECT_TEXTSELECTIONCHANGED = 0x8014
    # Sent when a window is uncloaked
    EVENT_OBJECT_UNCLOAKED = 0x8018
    EVENT_OBJECT_VALUECHANGE = 0x800E

    # The range of event constant values reserved for OEMs. For more information, see Allocation of WinEvent IDs.
    EVENT_OEM_DEFINED_START = 0x0101
    EVENT_OEM_DEFINED_END = 0x01FF

    EVENT_SYSTEM_ALERT = 0x0002
    EVENT_SYSTEM_ARRANGMENTPREVIEW = 0x8016
    EVENT_SYSTEM_CAPTUREEND = 0x0009
    EVENT_SYSTEM_CAPTURESTART = 0x0008
    EVENT_SYSTEM_CONTEXTHELPEND = 0x000D
    EVENT_SYSTEM_CONTEXTHELPSTART = 0x000C
    # The active desktop has been switched.
    EVENT_SYSTEM_DESKTOPSWITCH = 0x0020
    EVENT_SYSTEM_DIALOGEND = 0x0011
    EVENT_SYSTEM_DIALOGSTART = 0x0010
    EVENT_SYSTEM_DRAGDROPEND = 0x000F
    EVENT_SYSTEM_DRAGDROPSTART = 0x000E
    EVENT_SYSTEM_END = 0x00FF
    EVENT_SYSTEM_FOREGROUND = 0x0003
    EVENT_SYSTEM_MENUPOPUPEND = 0x0007
    EVENT_SYSTEM_MENUPOPUPSTART = 0x0006
    EVENT_SYSTEM_MENUEND = 0x0005
    EVENT_SYSTEM_MENUSTART = 0x0004
    # A window object is about to be restored
    EVENT_SYSTEM_MINIMIZEEND = 0x0017
    # A window object is about to be minimized
    EVENT_SYSTEM_MINIMIZESTART = 0x0016
    # The movement or resizing of a window has finished
    EVENT_SYSTEM_MOVESIZEEND = 0x000B
    EVENT_SYSTEM_MOVESIZESTART = 0x000A
    EVENT_SYSTEM_SCROLLINGEND = 0x0013
    EVENT_SYSTEM_SCROLLINGSTART = 0x0012
    EVENT_SYSTEM_SOUND = 0x0001
    EVENT_SYSTEM_SWITCHEND = 0x0015
    EVENT_SYSTEM_SWITCHSTART = 0x0014

    # The range of event constant values reserved for UI Automation event identifiers. For more information, see Allocation of WinEvent IDs.
    EVENT_UIA_EVENTID_START = 0x4E00
    EVENT_UIA_EVENTID_END = 0x4EFF

    # The range of event constant values reserved for UI Automation property-changed event identifiers. For more information, see Allocation of WinEvent IDs.
    EVENT_UIA_PROPID_START = 0x7500
    EVENT_UIA_PROPID_END = 0x75FF

    UNKNOWN = 0x00

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN



import subprocess
import time
from ctypes import *
from ctypes.wintypes import *
from tkinter import *
from tkinter import ttk
from typing import Optional

SCREEN_NAME = "virtual desktop stub"

all_process = []

user32 = WinDLL("user32", use_last_error=True)


def find_virtdeskstub() -> Optional[HWND]:
    hwnds = []

    buff = create_unicode_buffer(100)

    @WINFUNCTYPE(BOOL, HWND, LPARAM)
    def enum_windows_proc(hwnd: HWND, lParam: LPARAM) -> BOOL:
        user32.GetWindowTextW(hwnd, buff, 100)
        if str(buff.value) == SCREEN_NAME:
            hwnds.append(hwnd)
            return False
        return True

    user32.EnumDesktopWindows(None, enum_windows_proc, 0)
    if hwnds:
        return hwnds[0]


def find_or_create_virtdeskstub() -> HWND:
    hwnd = find_virtdeskstub()
    if hwnd:
        return hwnd
    global all_process
    proc = subprocess.Popen([r"py", __file__, "spawn"])
    all_process.append(proc)
    for _ in range(5):
        time.sleep(0.1)
        hwnd = find_virtdeskstub()
        if hwnd:
            break
    return hwnd, proc


def new_virtdeskstub():
    root = Tk(className=SCREEN_NAME)
    root.geometry("120x50+5+1390")
    root.resizable(width=False, height=False)
    frm = ttk.Frame(root, padding=5)
    frm.grid()
    # ttk.Label(
    #     frm,
    #     text="JigsawWM",
    # ).grid(column=0, row=0)
    ttk.Button(frm, text="Quit", command=root.destroy).grid(column=0, row=1)
    # root.overrideredirect(True)
    root.call("wm", "attributes", ".", "-topmost", "1")
    # root.withdraw()
    mainloop()


def kill_all():
    global all_process
    for p in all_process:
        p.kill()


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1 and sys.argv[1] == "spawn":
        # print(r"py .\jigsawwm\virtdeskstub.py spawn")
        new_virtdeskstub()
    else:
        # print("py -m jigsawwm.virtdeskstub")
        from jigsawwm.w32.ivirtualdesktopmanager import virtual_desktop_manager
        from jigsawwm.w32.window import Window, inspect_window

        hwnd = find_or_create_virtdeskstub()
        print("hwnd", hwnd)
        if hwnd:
            inspect_window(hwnd)
            try:
                virtual_desktop_manager.GetWindowDesktopId(hwnd)
            except:
                print()
                print("FAILED")
                print("FAILED")
                print("FAILED")
                print("FAILED")

        kill_all()



import enum
from ctypes import *
from ctypes.wintypes import *
from typing import Iterator, List, Tuple

user32 = WinDLL("user32", use_last_error=True)
shcore = WinDLL("shcore", use_last_error=True)
_current_pos_ptr = POINT()

# Ref: https://learn.microsoft.com/en-us/windows/win32/gdi/multiple-display-monitors-functions


def get_cursor_pos() -> POINT:
    """Retrieves the position of the mouse cursor, in screen coordinates.

    :return: mouse position
    :rtype: POINT
    """
    if not user32.GetCursorPos(pointer(_current_pos_ptr)):
        raise Exception("failed to get cursor position")
    return _current_pos_ptr


def set_cursor_pos(x: int, y: int):
    """Moves the cursor to the specified screen coordinates

    :param x: int
    :param y: int
    """
    if not user32.SetCursorPos(int(x), int(y)):
        raise Exception("failed to set cursor position")


def enum_display_monitors() -> List[HMONITOR]:
    """Returns a List of all monitors. THIS DO NOT RETURN MIRRORING MONITORS

    :return: list of monitor handles
    :rtype: List[]
    """
    hmons = []

    @WINFUNCTYPE(BOOL, HMONITOR, HDC, LPRECT, LPARAM)
    def monitor_enum_proc(
        hmon: HMONITOR, hdc: HDC, lprc: LPRECT, lParam: LPARAM
    ) -> BOOL:
        hmons.append(hmon)
        return True

    if not user32.EnumDisplayMonitors(None, None, monitor_enum_proc, None):
        raise WinError(get_last_error())
    return hmons


def monitor_from_point(x: int, y: int) -> HMONITOR:
    """Retrieves monitor from the specified coordinate
    
    :param int x: X
    :param int y: Y
    :returns: monitor handle
    :rtype: HMONITOR
    """
    return user32.MonitorFromPoint(POINT(x=x, y=y), 0)


def monitor_from_window(hwnd: HWND) -> HMONITOR:
    """Retrieves monitor from the specified window
    
    :param HWND hwn: window handle
    :returns: monitor handle
    :rtype: HMONITOR
    """
    return user32.MonitorFromWindow(hwnd, 0)


CCHDEVICENAME = 32


class MONITORINFOEX(Structure):
    cbSize: int
    rcMonitor: RECT
    rcWork: RECT
    dwFlags: int
    szDevice: CHAR * CCHDEVICENAME

    _fields_ = (
        ("cbSize", DWORD),
        ("rcMonitor", RECT),
        ("rcWork", RECT),
        ("dwFlags", DWORD),
        ("szDevice", CHAR * CCHDEVICENAME),
    )


class DEVICE_SCALE_FACTOR(enum.IntEnum):
    DEVICE_SCALE_FACTOR_INVALID = 0
    SCALE_100_PERCENT = 100
    SCALE_120_PERCENT = 120
    SCALE_125_PERCENT = 125
    SCALE_140_PERCENT = 140
    SCALE_150_PERCENT = 150
    SCALE_160_PERCENT = 160
    SCALE_175_PERCENT = 175
    SCALE_180_PERCENT = 180
    SCALE_200_PERCENT = 200
    SCALE_225_PERCENT = 225
    SCALE_250_PERCENT = 250
    SCALE_300_PERCENT = 300
    SCALE_350_PERCENT = 350
    SCALE_400_PERCENT = 400
    SCALE_450_PERCENT = 450
    SCALE_500_PERCENT = 500


class Monitor:
    """Represents a Display Monitor

    :param hmon: HMONITOR the monitor handle
    """

    _hmon: HMONITOR

    def __init__(self, hmon: HMONITOR):
        self._hmon = hmon

    def __eq__(self, other):
        return isinstance(other, Monitor) and self._hmon == other._hmon

    def __hash__(self):
        return hash(self._hmon)

    def get_info(self) -> MONITORINFOEX:
        """Retrieves monitor information
        
        :returns: monitor information
        :rtype: MONITORINFOEX
        """
        monitor_info = MONITORINFOEX()
        monitor_info.cbSize = sizeof(monitor_info)
        if not user32.GetMonitorInfoA(self._hmon, pointer(monitor_info)):
            raise WinError(get_last_error())
        return monitor_info

    def get_scale_factor(self) -> DEVICE_SCALE_FACTOR:
        """Retrieves monitor scale factor
        
        :returns: scale factor
        :rtype: DEVICE_SCALE_FACTOR
        """
        scale_factor = ULONG()
        if shcore.GetScaleFactorForMonitor(self._hmon, byref(scale_factor)) != 0:
            raise WinError(get_last_error())
        return DEVICE_SCALE_FACTOR(scale_factor.value)


def get_monitors() -> Iterator[Monitor]:
    """Retrieves all display monitors(mirroring monitors are excluded)

    :returns: system monitors
    :rtype: Iterator[Monitor]
    """
    return map(Monitor, enum_display_monitors())


def get_monitor_central(monitor: Monitor) -> Tuple[int, int]:
    """Retrieves coordinates of the center of specified monitor
    
    :param Monitor monitor: monitor
    :returns: X/Y coorindates
    :rtype: Tuple[int, int]
    """
    rect = monitor.get_info().rcMonitor
    return (
        rect.left + (rect.right - rect.left) / 2,
        rect.top + (rect.bottom - rect.top) / 2,
    )


def get_topo_sorted_monitors() -> List[Monitor]:
    """Sort monitor from left to right, top to bottom by Central Points
    
    :returns: list of monitors ordered by their central point, X and then Y
    :rtype: List[Monitor]
    """
    return sorted(get_monitors(), key=get_monitor_central)


def get_monitor_from_point(x: int, y: int) -> Monitor:
    """Retrieves monitor from X/Y coordinates
    
    :param int x: X coordinate
    :param int y: Y coordinate
    :returns: Monitor from specified point
    :rtype: Monitor
    """
    return Monitor(monitor_from_point(x, y))


def get_monitor_from_cursor() -> Monitor:
    """Retrieves monitor from cursor
    
    :returns: Monitor from current cursor
    :rtype: Monitor
    """
    pt = get_cursor_pos()
    return Monitor(monitor_from_point(pt.x, pt.y))


def get_monitor_from_window(hwnd: HWND) -> Monitor:
    """Retrieves monitor from window handle
    
    :param HWND hwnd: window handle
    :returns: Monitor that owns specified window
    :rtype: Monitor
    """
    if hwnd:
        return Monitor(monitor_from_window(hwnd))


if __name__ == "__main__":
    for monitor in get_topo_sorted_monitors():
        print("scale factor     :", monitor.get_scale_factor())
        monitor_info = monitor.get_info()
        print("device           :", monitor_info.szDevice)
        m = monitor_info.rcMonitor
        print("monitor rect     :", m.left, m.top, m.right, m.bottom)
        m = monitor_info.rcWork
        print("monitor workarea :", m.left, m.top, m.right, m.bottom)



import sys
import time
from ctypes import *
from ctypes.wintypes import *
from dataclasses import dataclass
from io import StringIO
from tkinter import messagebox
from typing import Callable, Iterator, List, Optional

from . import process
from .sendinput import *
from .vk import Vk
from .window_structs import *

user32 = WinDLL("user32", use_last_error=True)
kernel32 = WinDLL("kernel32", use_last_error=True)
dwmapi = WinDLL("dwmapi", use_last_error=True)


def enum_windows(
    check: Optional[Callable[[HWND], EnumCheckResult]] = None
) -> List[HWND]:
    """Returns a List of all top-level windows on current desktop.

    :param check: optional, to determinate if a HWN should be added to list, or stop iteration

    :return: list of window handles
    :rtype: List[HWND]
    """
    check = check or (lambda _: 1)
    hwnds = []

    @WINFUNCTYPE(BOOL, HWND, LPARAM)
    def enum_windows_proc(hwnd: HWND, lParam: LPARAM) -> BOOL:
        r = check(hwnd)
        if EnumCheckResult.CAPTURE in r:
            hwnds.append(hwnd)
        return EnumCheckResult.STOP not in r

    if not user32.EnumWindows(enum_windows_proc, None):
        last_error = get_last_error()
        if last_error:
            raise WinError(last_error)
    return hwnds


def get_foreground_window() -> HWND:
    """Retrieves foreground window handle"""
    return user32.GetForegroundWindow()


def get_window_style(hwnd: HWND) -> WindowStyle:
    """Retrieves style of the specified window handle"""
    return WindowStyle(user32.GetWindowLongA(hwnd, -16))


def get_window_exstyle(hwnd: HWND) -> WindowExStyle:
    """Retrieves ex-style of the specified window handle"""
    return WindowExStyle(user32.GetWindowLongA(hwnd, -20))


def show_window(hwnd: HWND, cmd: ShowWindowCmd):
    """Show window"""
    user32.ShowWindow(hwnd, cmd)


def minimize_window(hwnd: HWND):
    """Minimize window"""
    show_window(hwnd, ShowWindowCmd.SW_MINIMIZE)


def maximize_window(hwnd: HWND):
    """Maximize window"""
    show_window(hwnd, ShowWindowCmd.SW_MAXIMIZE)


def restore_window(hwnd: HWND):
    """Restore window"""
    show_window(hwnd, ShowWindowCmd.SW_RESTORE)


def get_window_title(hwnd: HWND) -> str:
    """Retrieves window title"""
    title = create_unicode_buffer(255)
    user32.GetWindowTextW(hwnd, title, 255)
    return str(title.value)


def get_window_class_name(hwnd: HWND) -> str:
    """Retrieves window class name"""
    buff = create_unicode_buffer(100)
    user32.GetClassNameW(hwnd, buff, 100)
    return str(buff.value)


def get_window_pid(hwnd: HWND) -> DWORD:
    """Retrieves id of the process that owns the window"""
    pid = DWORD()
    user32.GetWindowThreadProcessId(hwnd, pointer(pid))
    return pid


def is_window_visible(hwnd: HWND) -> bool:
    """Check if window is visible"""
    return bool(user32.IsWindowVisible(hwnd))


def is_window_cloaked(hwnd: HWND) -> bool:
    """Check if window is cloaked"""
    cloaked = INT()
    windll.dwmapi.DwmGetWindowAttribute(
        hwnd,
        DwmWindowAttribute.DWMWA_CLOAKED,
        pointer(cloaked),
        sizeof(cloaked),
    )
    return bool(cloaked.value)


def is_window(hwnd: HWND) -> bool:
    """Check if handle is a window handle"""
    return user32.IsWindow(hwnd)


def is_top_level_window(hwnd: HWND) -> bool:
    """Check if window is a top-level window"""
    return user32.IsTopLevelWindow(hwnd)


def is_app_window(hwnd: HWND, style: Optional[WindowExStyle] = None) -> bool:
    """Check if window is a app window (user mode / visible / resizable)"""
    style = style or get_window_style(hwnd)
    return bool(
        not is_window_cloaked(hwnd)
        and WindowStyle.SIZEBOX in style
        and not process.is_elevated(get_window_pid(hwnd))
    )


def is_manageable_window(
    hwnd: HWND,
    is_force_managed: Optional[Callable[[HWND], bool]] = None,
) -> bool:
    """Check if window is able to be managed by us"""
    is_force_managed = is_force_managed or (lambda hwnd: False)
    #  or is_force_managed(hwnd)
    style = get_window_style(hwnd)
    return bool(
        is_app_window(hwnd, style)
        and (get_window_title(hwnd) or is_force_managed(hwnd))
        and WindowStyle.MAXIMIZEBOX & style
        and WindowStyle.MINIMIZEBOX & style
        and WindowStyle.VISIBLE in style
        and not WindowStyle.MINIMIZE & style
    )


def get_first_app_window() -> HWND:
    """Retrieves any window from current desktop"""
    hwnd = get_foreground_window()
    if is_app_window(hwnd):
        return hwnd

    hwnds = enum_windows(
        lambda hwnd: EnumCheckResult.CAPTURE_AND_STOP
        if is_app_window(hwnd)
        else EnumCheckResult.SKIP
    )

    if hwnds:
        return hwnds[0]


def get_window_extended_frame_bounds(hwnd: HWND) -> RECT:
    """Retrieve extended frame bounds of the specified window"""
    bound = RECT()
    windll.dwmapi.DwmGetWindowAttribute(
        hwnd,
        DwmWindowAttribute.DWMWA_EXTENDED_FRAME_BOUNDS,
        pointer(bound),
        sizeof(bound),
    )
    return bound


def get_window_rect(hwnd: HWND) -> RECT:
    """Retrieves rect(position/size) of the specified window"""
    rect = RECT()
    if not user32.GetWindowRect(hwnd, pointer(rect)):
        raise WinError(get_last_error())
    return rect


SWP_NOACTIVATE = 0x0010
SET_WINDOW_RECT_FLAG = SWP_NOACTIVATE


def set_window_rect(hwnd: HWND, rect: RECT):
    """Move/resize specified window"""
    x, y, w, h = rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top
    if not user32.SetWindowPos(hwnd, None, x, y, w, h, SET_WINDOW_RECT_FLAG):
        raise WinError(get_last_error())


@dataclass
class Window:
    """Represents a top-level window

    :param hwnd: HWND the window handle
    """

    _hwnd: HWND
    _last_rect = None

    def __init__(self, hwnd: HWND):
        self._hwnd = hwnd

    def __eq__(self, other):
        return isinstance(other, Window) and self._hwnd == other._hwnd

    def __hash__(self):
        return hash(self._hwnd)

    @property
    def handle(self) -> HWND:
        return self._hwnd

    @property
    def title(self) -> str:
        """Retrieves the text of the specified window's title bar (if it has one)

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getwindowtexta

        :return: text of the title bar
        :rtype: str
        """
        return get_window_title(self._hwnd)

    @property
    def class_name(self):
        """Retrieves the name of the class to which the specified window belongs.

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getclassnamea

        :return: class name
        :rtype: str
        """
        return get_window_class_name(self._hwnd)

    @property
    def exe(self):
        """Retrieves the full path of the executable

        :return: full path of the executable
        :rtype: str
        """
        return process.get_exepath(self.pid)

    @property
    def pid(self) -> int:
        """Retrieves the process id

        :return: process id
        :rtype: int
        """
        return get_window_pid(self._hwnd)

    @property
    def is_visible(self) -> bool:
        """Determines the visibility state of the specified window.

        :return: If the specified window, its parent window, its parent's parent window,
            and so forth, have the WS_VISIBLE style, the return value is `True`.
            Otherwise, the return value is `False`.
        :rtype: bool
        """
        return is_window_visible(self._hwnd)

    def get_style(self) -> WindowStyle:
        """Retrieves style

        :return: window style
        :rtype: WindowStyle
        """
        return get_window_style(self._hwnd)

    def get_exstyle(self) -> WindowExStyle:
        """Retrieves ex-style

        :return: window ex-style
        :rtype: ExWindowStyle
        """
        return get_window_exstyle(self._hwnd)

    def minimize(self):
        """Minimizes the specified window and activates the next top-level window in the Z order."""
        minimize_window(self._hwnd)

    def maximize(self):
        """Activates the window and displays it as a maximized window."""
        maximize_window(self._hwnd)

    def restore(self):
        """Activates and displays the window. If the window is minimized or maximized,
        the system restores it to its original size and position."""
        restore_window(self._hwnd)

    def toggle_maximize(self):
        """Toggle maximize style"""
        if self.get_style() & WindowStyle.MAXIMIZE:
            self.restore()
        else:
            self.maximize()

    @property
    def is_evelated(self):
        """Check if window is elevated (Administrator)"""
        return process.is_elevated(self.pid)

    @property
    def is_cloaked(self) -> bool:
        """Check if window is cloaked (DWM)

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/dwmapi/ne-dwmapi-dwmwindowattribute
        """
        return is_window_cloaked(self._hwnd)

    def exists(self) -> bool:
        return is_window(self._hwnd)

    def get_extended_frame_bounds(self) -> RECT:
        """Retrieves extended frame bounds

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/dwmapi/ne-dwmapi-dwmwindowattribute
        """
        return get_window_extended_frame_bounds(self._hwnd)

    def get_rect(self) -> RECT:
        """Retrieves the dimensions of the bounding rectangle of the specified window.
        The dimensions are given in screen coordinates that are relative to the upper-left
        corner of the screen

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getwindowrect

        :return: a RECT with top/left/bottom/right properties
        :rtype: RECT
        """
        return get_window_rect(self._hwnd)

    def set_rect(self, rect: RECT):
        """Sets the dimensions of the bounding rectangle (Call SetWindowPos with RECT)

        Ref: https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setwindowpos

        :param rect: RECT with top/left/bottom/right properties
        """
        set_window_rect(self._hwnd, rect)
        self._last_rect = rect

    def activate(self) -> bool:
        """Brings the thread that created current window into the foreground and activates the window"""
        return set_active_window(self)

    @property
    def last_rect(self) -> Optional[RECT]:
        return self._last_rect


def get_app_windows() -> Iterator[Window]:
    """Get all manageable windows of specified/current desktop"""
    return map(
        Window,
        enum_windows(
            lambda hwnd: EnumCheckResult.CAPTURE
            if is_app_window(hwnd)
            else EnumCheckResult.SKIP
        ),
    )


def get_manageable_windows(
    is_force_managed: Optional[Callable[[HWND], bool]] = None,
) -> Iterator[Window]:
    """Get all manageable windows of specified/current desktop"""
    return map(
        Window,
        enum_windows(
            lambda hwnd: EnumCheckResult.CAPTURE
            if is_manageable_window(hwnd, is_force_managed)
            else EnumCheckResult.SKIP
        ),
    )


def get_active_window() -> Optional[Window]:
    """Retrieves current activated window"""
    hwnd = get_foreground_window()
    if hwnd:
        return Window(hwnd)


def set_active_window(window: Window) -> bool:
    """Brings the thread that created the specified window into the foreground and activates the window

    Ref: https://github.com/AutoHotkey/AutoHotkey/blob/e379b60e44d35494d4a19d1e5001f2dd38773391/source/window.cpp#L25
    """
    # simple way
    if user32.SetForegroundWindow(window.handle):
        # print("simple way works")
        return
    # well, simple way didn't work, we have to make our process Foreground
    our_thread_id = kernel32.GetCurrentThreadId()
    fore_thread_id = None
    target_thread_id = user32.GetWindowThreadProcessId(window.handle, None)

    uf = False  # attached our thread to the fore thread
    ft = False  # attached the fore thread to the target thread
    curr_fore_hwnd = user32.GetForegroundWindow()
    if curr_fore_hwnd:
        fore_thread_id = user32.GetWindowThreadProcessId(curr_fore_hwnd, None)
        if fore_thread_id and fore_thread_id != our_thread_id:
            uf = user32.AttachThreadInput(our_thread_id, fore_thread_id, True)
            # print("attach our thread to the fore thread:", uf)
        if fore_thread_id and target_thread_id and fore_thread_id != target_thread_id:
            ft = user32.AttachThreadInput(fore_thread_id, target_thread_id, True)
            # print("attach fore thread to the target thread:", ft)
    new_fore_window = None
    retry = 5
    while new_fore_window != window.handle and retry > 0:
        send_input(
            INPUT(
                type=INPUTTYPE.KEYBOARD,
                ki=KEYBDINPUT(wVk=Vk.MENU, dwFlags=KEYEVENTF.KEYUP),
            ),
            INPUT(
                type=INPUTTYPE.KEYBOARD,
                ki=KEYBDINPUT(wVk=Vk.MENU, dwFlags=KEYEVENTF.KEYUP),
            ),
        )
        user32.SetForegroundWindow(window.handle)
        new_fore_window = user32.GetForegroundWindow()
        retry -= 1
        time.sleep(0.01)
    # print(
    #     f"our: {our_thread_id}   fore: {fore_thread_id}   target{target_thread_id}  succeeded: {new_fore_window == window.handle}"
    # )
    # detach input thread
    if uf:
        user32.AttachThreadInput(our_thread_id, fore_thread_id, False)
    if ft:
        user32.AttachThreadInput(fore_thread_id, target_thread_id, False)
    # print("detached")


def minimize_active_window():
    """Minize active window"""
    window = get_active_window()
    if window:
        window.minimize()


def toggle_maximize_active_window():
    """Maximize/Unmaximize active window"""
    window = get_active_window()
    if window:
        window.toggle_maximize()


def sprint_window(hwnd: HWND) -> str:
    f = StringIO()
    inspect_window(hwnd, file=f)
    return f.getvalue()


def inspect_window(hwnd: HWND, file=sys.stdout):
    print(file=file)
    window = Window(hwnd)
    if not window.exists():
        print("window doesn't exist anymore")
        return
    print("hwnd         :", window.handle, file=file)
    print("title        :", window.title, file=file)
    print("pid          :", window.pid, file=file)
    print("class name   :", window.class_name, file=file)
    print("exe path     :", window.exe, file=file)
    style = window.get_style()
    style_flags = []
    for s in WindowStyle:
        if s in style:
            style_flags.append(s.name)
    print("style        :", ", ".join(style_flags), file=file)
    exstyle = window.get_exstyle()
    exstyle_flags = []
    for s in WindowExStyle:
        if s in exstyle:
            exstyle_flags.append(s.name)
    print("exstyle      :", ", ".join(exstyle_flags), file=file)
    print("is_cloaked   :", window.is_cloaked, file=file)
    rect = window.get_rect()
    print("rect         :", rect.left, rect.top, rect.right, rect.bottom, file=file)
    bound = window.get_extended_frame_bounds()
    print("bound        :", bound.left, bound.top, bound.right, bound.bottom, file=file)
    print("is_app_window:", is_app_window(hwnd), file=file)
    print("is_manageable:", is_manageable_window(hwnd), file=file)
    print("is_evelated  :", window.is_evelated, file=file)


def inspect_active_window():
    text = sprint_window(get_foreground_window())
    messagebox.showinfo("JigsawWM", text)


if __name__ == "__main__":
    import time

    time.sleep(2)
    inspect_active_window()
    # for window in get_app_windows():
    #     inspect_window(window.handle)
    # for win in get_windows():
    #     inspect_window(win)



import abc
import logging
import os
import signal
import sys
import traceback
from subprocess import PIPE, Popen
from threading import Lock
from tkinter import messagebox
from typing import Callable, List, Sequence, TextIO

import pystray
from PIL import Image

from jigsawwm import jmk

# support for Ctrl+C in console
signal.signal(signal.SIGINT, signal.SIG_DFL)

logger = logging.getLogger(__name__)


# show exception both CLI/GUI
def handle_exc(exc_type, exc_value, exc_traceback):
    msg = traceback.format_exception(exc_type, exc_value, exc_traceback)
    print(msg, file=sys.stderr)
    messagebox.showerror("JigsawWM", msg)


jmk.handle_exc = handle_exc


class Job(abc.ABC):
    autorun = True

    """Job is the fundamental building block of the Daemon, represent a Unit
    of Automation.

    .. code-block::
                        Job
                        /  \
            Service(wm)     Task(daily_routine: open websites, workday_routine: launch thunderbird, im)
                /
        ProcessService(syncthing)
    """

    @abc.abstractproperty
    def name(self):
        pass

    @abc.abstractproperty
    def text(self):
        pass

    @abc.abstractmethod
    def launch(self):
        pass


class Service(Job):
    """Service is a kind of Job represents the longlived automation.
    users register arbitrary service to the Daemon and turn them on or off as they pleased
    """

    @abc.abstractproperty
    def is_running(self) -> bool:
        pass

    @abc.abstractmethod
    def start(self):
        pass

    @abc.abstractmethod
    def stop(self):
        pass

    def launch(self):
        self.start()

    def toggle(self):
        self.stop() if self.is_running else self.start()

    def restart(self):
        self.stop()
        self.start()

    @property
    def text(self):
        status = "running" if self.is_running else "stopped"
        return f"[{status}] {self.name}"


class ProcessService(Service):
    """ProcessService is a kind of specialized Service that run a CLI program in the
    background"""

    log_path: str
    log_append_only: bool = False
    _process: Popen = None
    _log_file: TextIO = None
    _lock: Lock

    def __init__(self):
        self._lock = Lock()

    @abc.abstractproperty
    def args(self) -> List[str]:
        pass

    def is_running(self) -> bool:
        with self._lock:
            if self._process is None:
                return False
            if self._process.poll() is None:
                return True
        return False

    def start(self):
        """Start the program in the background"""
        with self._lock:
            if self._process is not None:
                raise ValueError(f"Service {self.name} is already running")
            log_file = PIPE
            if self.log_path:
                log_file = open(self.log_path, "a+" if self.log_append_only else "w+")
                self._log_file = log_file
            self._process = Popen(
                self.args, stdout=log_file, stdin=log_file, shell=True
            )

    def stop(self):
        """Stop the program"""
        with self._lock:
            if self._process is None:
                return
            Popen("TASKKILL /F /PID {pid} /T".format(pid=self._process.pid))
            self._process = None
            if self._log_file:
                self._log_file.close()
            self._log_file = None


class Task(Job):
    """Task is a shortlived automation in constrast to Service"""

    def condition(self) -> bool:
        return True

    @abc.abstractmethod
    def run(self) -> bool:
        pass

    def launch(self):
        if self.condition():
            self.run()

    @property
    def text(self):
        return self.name


class Daemon:
    """JigsawWM Daemon serivce, you must inherite this class and override the `setup` function
    to configurate the Manager.
    """

    trayicon: pystray.Icon = None
    jobs: Sequence[Job] = []

    def __init__(self):
        jmk.handle_exc = handle_exc

    def create_trayicon(self):
        """Start trayicon"""
        script_dir = os.path.dirname(__file__)
        icon_path = os.path.join(script_dir, "assets", "logo.png")
        icon = Image.open(icon_path)

        tray_icon = pystray.Icon(
            "JigsawWM",
            icon=icon,
            menu=pystray.Menu(self.tray_menu_items),
        )
        self.trayicon = tray_icon
        self.trayicon.run()

    def tray_menu_items(self) -> List[pystray.MenuItem]:
        service_menu_items = [
            pystray.MenuItem(
                service.text,
                service.toggle,
            )
            for service in self.jobs
            if isinstance(service, Service)
        ]
        task_menu_items = [
            pystray.MenuItem(
                task.text,
                task.run,
            )
            for task in self.jobs
            if isinstance(task, Task)
        ]
        return [
            *task_menu_items,
            pystray.Menu.SEPARATOR,
            *service_menu_items,
            pystray.Menu.SEPARATOR,
            pystray.MenuItem("Exit", self.stop),
        ]

    def stop(self):
        """Stop daemon service"""
        logger.info(f"stopping daemon")
        self.trayicon.stop()
        self.trayicon = None
        signal.raise_signal(signal.SIGINT)

    def register(self, job: Callable[[], Job]):
        logger.info(f"registering {job.name}")
        job: Job = job()
        if job.autorun:
            logger.info(f"autorun {job.name}")
            job.launch()
        self.jobs.append(job)

    def message_loop(self):
        logger.info(f"start message loop")
        self.create_trayicon()


if os.environ.get("DEBUG"):
    logging.basicConfig(level=logging.DEBUG)
instance = Daemon()
register = instance.register
message_loop = instance.message_loop
stop = instance.stop
